# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos sobjects tests"""

from cubicweb import Unauthorized
from cubicweb.devtools.testlib import CubicWebTC

from unittest_sobjects import silent_logging


class SKOSSchemaTC(CubicWebTC):

    def test_concept_label_no_language(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'scheme')
            concept = scheme.add_concept(u'a concept', language_code=None)
            cnx.commit()
            concept.cw_clear_all_caches()
            self.assertEqual(concept.labels, {None: 'a concept'})
            self.assertEqual(concept.label(), 'a concept')

    def test_scheme_no_title(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme')
            cnx.commit()
            # title default to cwuri
            self.assertEqual(scheme.dc_title(), 'http://testing.fr/cubicweb/%s' % scheme.eid)

    def test_concept_multiple_schemes(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme1 = cnx.create_entity('ConceptScheme', title=u'some scheme')
            concept = scheme1.add_concept(u'a concept', language_code=u'en')
            scheme2 = cnx.create_entity('ConceptScheme', title=u'some other scheme')
            concept.cw_set(in_scheme=scheme2)
            cnx.commit()
            self.assertEqual(set(x.eid for x in concept.in_scheme),
                             set([scheme1.eid, scheme2.eid]))
            # "scheme" property picks a scheme randomly, as we don't want to clutter the code for
            # this peculiarity (do we?)
            self.assertIn(concept.scheme.eid, [scheme1.eid, scheme2.eid])

    def test_concept_multiple_parents(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'some scheme')
            concept1 = scheme.add_concept(u'a concept', language_code=u'en')
            concept2 = scheme.add_concept(u'another concept', language_code=u'en')
            subconcept = concept1.add_concept(u'a sub-concept', language_code=u'en')
            subconcept.cw_set(broader_concept=concept2)
            cnx.commit()
            self.assertEqual(set(x.eid for x in subconcept.broader_concept),
                             set([concept1.eid, concept2.eid]))
            # "parent_concept" property picks a scheme randomly, as we don't want to clutter the
            # code for this peculiarity (do we?)
            self.assertIn(subconcept.parent_concept.eid, [concept1.eid, concept2.eid])

    def test_in_scheme_permissions(self):
        with self.admin_access.repo_cnx() as cnx:
            url = u'file://%s' % self.datapath('siaf_matieres_shortened.xml')
            cnx.create_entity('CWSource', name=u'mythesaurus',
                              type=u'datafeed', parser=u'rdf.skos', url=url)
            cnx.commit()
            dfsource = self.repo.sources_by_uri[u'mythesaurus']
            with self.repo.internal_cnx() as icnx:
                with silent_logging('cubicweb.sources'):
                    stats = dfsource.pull_data(icnx, force=True, raise_on_error=False)
            assert stats['created']
            scheme = cnx.find('ConceptScheme').one()
            with self.assertRaises(Unauthorized) as cm:
                cnx.execute('DELETE X in_scheme S WHERE X is Concept, '
                            'S cw_source SO, SO name "mythesaurus"')
                cnx.commit()
            cnx.rollback()
            self.assertEqual(cm.exception.args,
                             ('delete', 'relation Concept in_scheme ConceptScheme'))
            with self.assertRaises(Unauthorized) as cm:
                cnx.create_entity('Concept', in_scheme=scheme)
                cnx.commit()
            cnx.rollback()
            self.assertEqual(cm.exception.args,
                             ('add', 'relation Concept in_scheme ConceptScheme'))


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos views"""

from copy import copy

from rql.utils import uquote

from cubicweb import ValidationError, tags

from cubicweb.predicates import (is_instance, has_related_entities, one_line_rset,
                                 score_entity, match_edited_type,
                                 match_form_params, match_user_groups,
                                 ExpectedValuePredicate)
from cubicweb.view import View, AnyRsetView, EntityView
from cubicweb.web import Redirect
from cubicweb.web import (action, component, formfields as ff, formwidgets as fw,
                          facet)
from cubicweb.web.views import (autoform, editcontroller, uicfg, ibreadcrumbs,
                                forms, tabs, cwsources, treeview)

from cubes.skos import to_unicode, lcsv
from cubes.skos.entities import CSVIndentationError, CSVDecodeError


_ = unicode

# uicfg tweaks #################################################################

# configure index page
uicfg.indexview_etype_section['Label'] = 'hidden'

pvs = uicfg.primaryview_section
# configure primary view for ConceptScheme
pvs.tag_attribute(('ConceptScheme', 'title'), 'hidden')
pvs.tag_attribute(('ConceptScheme', 'cwuri'), 'attributes')

pvs.tag_object_of(('Concept', 'in_scheme', 'ConceptScheme'), 'hidden')
# configure primary view for Concept
pvs.tag_attribute(('Concept', 'cwuri'), 'attributes')
for label_type in ('preferred', 'alternative', 'hidden'):
    pvs.tag_subject_of(('*', '%s_label' % label_type, '*'), 'hidden')
pvs.tag_object_of(('*', 'label_of', '*'), 'hidden')
pvs.tag_subject_of(('Concept', 'in_scheme', '*'), 'hidden')
pvs.tag_subject_of(('Concept', 'broader_concept', 'Concept'), 'hidden')
pvs.tag_object_of(('Concept', 'broader_concept', 'Concept'), 'hidden')
pvs.tag_subject_of(('Concept', 'narrower_concept', 'Concept'), 'hidden')
pvs.tag_object_of(('Concept', 'narrower_concept', 'Concept'), 'hidden')

pvs.tag_subject_of(('SKOSSource', 'through_cw_source', '*'), 'hidden')

afs = uicfg.autoform_section
affk = uicfg.autoform_field_kwargs
# configure creation/edit for ConceptScheme
afs.tag_object_of(('*', 'in_scheme', '*'), 'main', 'hidden')
affk.set_field_kwargs('ConceptScheme', 'title', widget=fw.TextInput({'size': 80}))
# configure creation/edit for Concept
afs.tag_subject_of(('*', 'in_scheme', '*'), 'main', 'hidden')
afs.tag_subject_of(('Concept', 'broader_concept', 'Concept'), 'main', 'relations')
afs.tag_object_of(('Concept', 'broader_concept', 'Concept'), 'main', 'hidden')
afs.tag_subject_of(('Concept', 'narrower_concept', 'Concept'), 'main', 'hidden')
afs.tag_object_of(('Concept', 'narrower_concept', 'Concept'), 'main', 'hidden')
for label_type in ('preferred', 'alternative', 'hidden'):
    afs.tag_subject_of(('*', '%s_label' % label_type, '*'), 'main', 'hidden')
afs.tag_object_of(('*', 'label_of', '*'), 'main', 'inlined')
affk.set_field_kwargs('Label', 'label', widget=fw.TextInput({'size': 80}))

# clickable cwuri in primary views
pvdc = uicfg.primaryview_display_ctrl
pvdc.tag_attribute(('*', 'cwuri'), {'vid': 'urlattr'})


class AddConceptToConceptAction(action.LinkToEntityAction):
    __abstract__ = True
    __select__ = action.LinkToEntityAction.__select__ & is_instance('Concept')
    target_etype = 'Concept'

    def url(self):
        baseurl = super(AddConceptToConceptAction, self).url()
        entity = self.cw_rset.get_entity(0, 0)
        linkto = 'in_scheme:%s:subject' % (entity.scheme.eid)
        return '%s&__linkto=%s' % (baseurl, self._cw.url_quote(linkto))


class AddNarrowerConceptAction(AddConceptToConceptAction):
    __regid__ = 'skos.add_broader_concept'
    title = _('add Concept broader_concept Concept object')
    rtype = 'broader_concept'
    role = 'object'


class AddRelatedConceptAction(AddConceptToConceptAction):
    __regid__ = 'skos.add_related_concept'
    title = _('add Concept related_concept Concept subject')
    rtype = 'related_concept'
    role = 'subject'


# base views ###################################################################


class ConceptSchemePrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('ConceptScheme')
    tabs = ['main_tab', _('skos_top_concepts_tab')]


class ConceptSchemeConceptsTab(tabs.TabsMixin, EntityView):
    """display a SKOS concept scheme tree"""
    __regid__ = 'skos_top_concepts_tab'  # don't use '.' in tab's regid
    __select__ = is_instance('ConceptScheme')

    def entity_call(self, entity):
        rschema = self._cw.vreg.schema.rschema('in_scheme')
        if rschema.has_perm(self._cw, 'add', toeid=entity.eid):
            self.w(tags.a(self._cw._('import concepts'),
                          href=entity.absolute_url(vid='skos.scheme.import'),
                          klass='btn btn-success pull-right'))
            self.w(tags.div(klass='clearfix'))
        rset = entity.top_concepts_rset
        if rset:
            treeid = 'skos_tree_%s' % entity.eid
            self._cw.view('treeview', rset=rset, treeid=treeid, initial_thru_ajax=True, w=self.w)


class ConceptPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('Concept')
    tabs = ['main_tab', _('skos_sub_concepts_tab')]


class ConceptConceptsTab(tabs.TabsMixin, EntityView):
    """display a SKOS concept tree"""
    __regid__ = 'skos_sub_concepts_tab'  # don't use '.' in tab's regid
    __select__ = (is_instance('Concept') &
                  has_related_entities('narrower_concept'))

    def entity_call(self, entity):
        rset = entity.related('narrower_concept')
        treeid = 'skos_tree_%s' % entity.eid
        self._cw.view('treeview', rset=rset, treeid=treeid, initial_thru_ajax=True, w=self.w)


class ConceptLabelsComponent(component.EntityCtxComponent):
    """display concept labels in a table"""
    __regid__ = 'skos.concept.labels'
    __select__ = component.EntityCtxComponent.__select__ & is_instance('Concept')
    title = _('Labels')
    context = 'navcontentbottom'
    order = 2

    def render_body(self, w):
        rql = ('Any LL,LLC,LK ORDERBY LLC WHERE L label_of X, X eid %(x)s, '
               'L language_code LLC, L label LL, L kind LK')
        rset = self._cw.execute(rql, {'x': self.entity.eid})
        self._cw.view('table', rset=rset, w=w)


class ConceptMultiSchemeComponent(component.EntityCtxComponent):
    """display a SKOS concept's schemes if there are more than one"""
    __regid__ = 'skos.concept.parent-schemes'
    __select__ = (component.EntityCtxComponent.__select__ &
                  is_instance('Concept') &
                  score_entity(lambda x: len(x.in_scheme) > 1))
    title = _('ConceptScheme_plural')
    context = 'navcontentbottom'
    order = 1

    def render_body(self, w):
        w(self._cw._('This concept belongs to several schemes:'))
        self._cw.view('list', rset=self.entity.related('in_scheme'))


class TranslateView(AnyRsetView):
    __regid__ = 'translate'

    def cell_call(self, row, col, props=None, format='text/html'):
        self.w(self._cw._(self.cw_rset.rows[row][col]))


class ConceptIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('Concept')

    def parent_entity(self):
        return self.entity.parent_concept or self.entity.scheme


# tree view ####################################################################

class ConceptTreeViewItemView(treeview.TreeViewItemView):
    """`treeitem` view for Concept with branch state closed by default"""
    __select__ = treeview.TreeViewItemView.__select__ & is_instance('Concept')

    def open_state(self, *args):
        return False


# facets #######################################################################

class InSchemeFacet(facet.RelationAttributeFacet):
    __regid__ = 'skos.facet.in_scheme'
    __select__ = facet.RelationAttributeFacet.__select__ & is_instance('Concept')
    rtype = 'in_scheme'
    role = 'subject'
    target_attr = 'title'


# ConceptScheme edition ########################################################

class ConceptSchemeAutoform(autoform.AutomaticEntityForm):
    """Autoform for ConceptScheme (creation only) with additional fields to
    import concepts from a file.
    """
    __select__ = (autoform.AutomaticEntityForm.__select__ & is_instance('ConceptScheme')
                  & score_entity(lambda x: not x.has_eid()))

    def __init__(self, *args, **kwargs):
        super(ConceptSchemeAutoform, self).__init__(*args, **kwargs)
        self.fieldsets_in_order = [None, _('inline csv import')]
        for field in ImportSchemeConceptsForm._fields_:
            field = copy(field)
            field.required = False
            field.fieldset = 'inline csv import'
            self.append_field(field)


class params_in_form(ExpectedValuePredicate):
    """Return 1 if expected values match with non-empty form parameters."""

    def __call__(self, cls, req, **kwargs):
        for param in self.expected:
            if not req.form.get(param):
                return 0
        return 1


class ConceptSchemeEditController(editcontroller.EditController):
    """Custom edit controller for ConceptScheme entity to handle import of
    concepts from file provided in edition form.
    """
    __select__ = (editcontroller.EditController.__select__ &
                  match_edited_type('ConceptScheme') &
                  match_form_params('stream', 'format', 'encoding', 'language_code', 'delimiter') &
                  params_in_form('stream'))

    def edit_entity(self, formparams, multiple=False):
        """Use created entity and import concepts from specified file."""
        eid = super(ConceptSchemeEditController, self).edit_entity(formparams, multiple=multiple)
        form = self._cw.form
        entity = self._cw.entity_from_eid(eid)
        _, fobj = form['stream']
        delimiter = CSV_DELIMITERS[form['delimiter']]
        _handle_concepts_import(self._cw, entity, form['format'],
                                stream=fobj, encoding=form['encoding'],
                                language_code=form['language_code'],
                                delimiter=delimiter)
        return eid


# simple file import ###########################################################

class ImportSchemeConceptsMixIn(object):
    __regid__ = 'skos.scheme.import'
    __select__ = (is_instance('ConceptScheme')
                  & match_user_groups('managers')
                  & one_line_rset())


class ImportSchemeConceptsForm(ImportSchemeConceptsMixIn, forms.EntityFieldsForm):
    """ask for a file to import"""
    filefield = ff.FileField(name='stream', label=_('file'), required=True)
    formatfield = ff.StringField(name='format', label=_('file format'),
                                 required=True, internationalizable=True,
                                 choices=[(_('simple'), u'simple'),
                                          (_(u'LCSV'), u'lcsv')], sort=False)
    encoding = ff.StringField(name='encoding', label=_('file encoding'),
                              internationalizable=True,
                              choices=[(_('utf-8'), u'utf-8'),
                                       (_('latin 1'), u'latin1')], sort=False)
    language = ff.StringField(name='language_code',
                              label=_('language of concepts in the file'),
                              internationalizable=True,
                              choices=[(_('french'), u'fr'),
                                       (_('english'), u'en')], sort=False)
    delimiter = ff.StringField(name='delimiter', label=_('hierarchical delimiter'),
                               internationalizable=True,
                               choices=[(_('tabulation'), 'tab'),
                                        (_(';'), ';'),
                                        (_(','), ','),
                                        (_('space'), 'space')], sort=False)
    form_buttons = [fw.SubmitButton(label=_('to_import'))]

    @property
    def action(self):
        return self.edited_entity.absolute_url(vid=self.__regid__)


CSV_DELIMITERS = {'tab': '\t', 'space': ' ', ',': ',', ';': ';'}


class ImportSchemeConceptsView(ImportSchemeConceptsMixIn, EntityView):

    def entity_call(self, entity):
        self.w('<h1>%s</h1>' % self._cw._('Import concepts from file'))
        form = self._cw.vreg['forms'].select(self.__regid__, self._cw, entity=entity)
        if form.posting:
            posted = form.process_posted()
            if 'stream' not in posted:  # https://www.cubicweb.org/ticket/5245936
                raise ValidationError(None, {'stream': self._cw.__("required field")})
            posted['delimiter'] = CSV_DELIMITERS[posted['delimiter']]
            _handle_concepts_import(self._cw, entity, posted.pop('format'), **posted)
            raise Redirect(entity.absolute_url(__message=self._cw._('Import completed')))
        else:
            form.render(w=self.w)


def _handle_concepts_import(req, entity, fileformat, **kwargs):
    """Entry point of import of concepts to be related to concept scheme
    `entity` based of specified file format.
    """
    assert fileformat in ('simple', 'lcsv'), fileformat
    if fileformat == 'simple':
        _simple_import(req, entity, **kwargs)
    else:
        _lcsv_import(req, entity, **kwargs)


def _simple_import(req, entity, **kwargs):
    """Trigger import of concepts to be related to concept scheme `entity`
    using a the 'simple' text format.
    """
    try:
        entity.add_concepts_from_file(**kwargs)
    except CSVIndentationError as exc:
        raise ValidationError(entity.eid,
                              {None: req._("Indentation error line %s") % exc.line})
    except CSVDecodeError as exc:
        raise ValidationError(entity.eid,
                              {None: req._("Encoding error line %s") % exc.line})


def _lcsv_import(req, entity, **kwargs):
    """Trigger import of concepts to be related to concept scheme `entity`
    using the LCSV text format.
    """
    try:
        req.cnx.call_service('lcsv.skos.import', scheme_uri=entity.cwuri, **kwargs)
    except lcsv.InvalidLCSVFile as exc:
        if exc.column:
            msg = req._(to_unicode(exc)) % exc.column
        else:
            msg = req._(to_unicode(exc))
        raise ValidationError(entity.eid, {None: msg})


# SKOS import ##################################################################

class ImportConceptSchemeAction(action.Action):
    __regid__ = 'rdf.skos.import'
    __select__ = match_user_groups('managers')
    title = _('import concept scheme from SKOS')
    category = 'manage'

    def url(self):
        return self._cw.build_url(vid=self.__regid__)


class SKOSImportForm(forms.FieldsForm):
    """ask for a file to import"""
    __regid__ = 'rdf.skos.import'
    __select__ = match_user_groups('managers')
    filefield = ff.FileField(name='file', label=_('SKOS file'), required=True)
    formatfield = ff.StringField(name='format', label=_('RDF format'),
                                 internationalizable=True,
                                 choices=[(_('per extension'), ''),
                                          (_('xml'), u'xml'),
                                          (_('ntriples'), u'nt'),
                                          (_('n3'), u'n3')],
                                 sort=False)
    form_buttons = [fw.SubmitButton(label=_('to_import'))]

    @property
    def action(self):
        return self._cw.build_url(vid=self.__regid__)


class SKOSImportResultView(View):
    __regid__ = 'rdf.skos.import'
    __select__ = match_user_groups('managers')

    def call(self):
        self.w('<h1>%s</h1>' % self._cw._('Import a SKOS file'))
        form = self._cw.vreg['forms'].select(self.__regid__, self._cw)
        if form.posting:
            posted = form.process_posted()
            if 'file' not in posted:  # https://www.cubicweb.org/ticket/5245936
                raise ValidationError(None, {'file': self._cw.__("required field")})
            stream = posted['file']
            rdf_format = posted['format']
            try:
                logs, scheme_uris = self._cw.cnx.call_service('rdf.skos.import', stream=stream,
                                                              rdf_format=rdf_format)
            except Exception as exc:
                logs = None
                msg = self._cw._('SKOS import failed: ') + to_unicode(exc)
                self.exception('error while importing %s', stream.filename)
            else:
                if scheme_uris:
                    schemes = self._cw.execute('Any X WHERE X is ConceptScheme, X cwuri IN (%s)'
                                               % ','.join(uquote(uri) for uri in scheme_uris))
                    msg = self._cw._('SKOS import completed: %s') % self._cw.view('csv', schemes)
                else:
                    msg = self._cw._('SKOS import completed but no scheme was found')
            self.w(u'<div class="message">%s</div>' % msg)
            if logs:
                self._cw.view('cw.log.table',
                              pyvalue=cwsources.log_to_table(self._cw, u''.join(logs)),
                              default_level='Info', w=self.w)

        else:
            form.render(w=self.w)

# SKOSSource primary view ##########################################################################


class SKOSSourcePrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('SKOSSource')
    tabs = [_('skos.source-main'), _('skos.source-imports')]
    default_tab = 'skos.source-main'


class SKOSSourceMainTab(tabs.PrimaryTab):
    __regid__ = 'skos.source-main'
    __select__ = is_instance('SKOSSource')


class SKOSSourceImportsTab(EntityView):
    __regid__ = 'skos.source-imports'
    __select__ = is_instance('SKOSSource')

    def entity_call(self, entity):
        if self._cw.user.is_in_group('managers'):
            self.w(tags.a(self._cw._('synchronize'), klass='btn btn-success pull-right',
                          href=entity.absolute_url(vid='skos.source-sync')))
            self.w(tags.div(klass='clearfix'))
        source = entity.through_cw_source[0]
        rset = source.related('cw_import_of', role='object')
        if rset:
            self._cw.view('cw.imports-table', rset=rset, w=self.w)


class SKOSSourceSyncView(EntityView):
    __regid__ = 'skos.source-sync'
    __select__ = (match_user_groups('managers')
                  & one_line_rset() & is_instance('SKOSSource'))

    title = _('synchronize')

    def entity_call(self, entity):
        self._cw.call_service('source-sync', source_eid=entity.through_cw_source[0].eid)
        msg = self._cw._('Source has been synchronized')
        url = entity.absolute_url(tab='skos.source-imports', __message=msg)
        raise Redirect(url)


# https://www.cubicweb.org/ticket/5474286 ##########################################################

from cubicweb.server import Service  # noqa


class CWSourceSyncAction(action.Action):
    __regid__ = 'cw.source-sync'
    __select__ = (action.Action.__select__ & match_user_groups('managers')
                  & one_line_rset() & is_instance('CWSource')
                  & score_entity(lambda x: x.name != 'system'))

    title = _('synchronize')
    category = 'mainactions'
    order = 20

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid=self.__regid__)


class CWSourceSyncView(EntityView):
    __regid__ = 'cw.source-sync'
    __select__ = (match_user_groups('managers')
                  & one_line_rset() & is_instance('CWSource')
                  & score_entity(lambda x: x.name != 'system'))

    title = _('synchronize')

    def entity_call(self, entity):
        self._cw.call_service('source-sync', source_eid=entity.eid)
        msg = self._cw._('Source has been synchronized')
        url = entity.absolute_url(tab='cwsource-imports', __message=msg)
        raise Redirect(url)


class SourceSynchronizationService(Service):
    """Force synchronization of a datafeed source"""
    __regid__ = 'source-sync'
    __select__ = Service.__select__ & match_user_groups('managers')

    def call(self, source_eid, **kwargs):
        source_entity = self._cw.entity_from_eid(source_eid)
        assert source_entity.name != 'system'
        repo = self._cw.repo  # Service are repo side only.
        dfsource = repo.sources_by_uri[source_entity.name]
        assert repo.config.source_enabled(dfsource)
        kwargs.setdefault('force', True)
        with repo.internal_cnx() as cnx:
            return dfsource.pull_data(cnx, **kwargs)



# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import os
from io import BytesIO
from tempfile import NamedTemporaryFile

from cubicweb.web import Redirect
from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos.rdfio import default_graph


class ViewsTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.create_entity('ConceptScheme', title=u'musique')
            cnx.commit()

    def test_flat_scheme_concepts_import(self):
        with self.admin_access.web_request() as req:
            scheme = req.find('ConceptScheme', title=u'musique').one()
            # simply test the form properly render and is well formed
            self.view('skos.scheme.import', rset=scheme.as_rset(), req=req, template=None)
            req.form = self.fake_form('skos.scheme.import', {
                'stream': ('filename.txt', BytesIO('\n\nélectro\nhip-hop\nrap\njazz\nclassique\n')),
                'encoding': u'utf-8',
                'language_code': u'fr',
                'format': u'simple',
                'delimiter': u'tab',
                }, [(scheme, {})])
            # now actually tests the import, using scheme.view and not self.view which doesn't like
            # exception, even Redirect
            self.assertRaises(Redirect, scheme.view, 'skos.scheme.import')
            self.assertEqual(set(c.dc_title() for c in scheme.top_concepts),
                             set(u'électro hip-hop rap jazz classique'.split()))
            self.assertEqual(set(l.language_code for c in scheme.top_concepts for l in c.preferred_label),
                             set(['fr']))

    def test_lcsv_scheme_concepts_import(self):
        with self.admin_access.web_request() as req:
            scheme = req.find('ConceptScheme', title=u'musique').one()
            # simply test the form properly render and is well formed
            self.view('skos.scheme.import', rset=scheme.as_rset(), req=req,
                      template=None)
            fname = 'lcsv_example_shortened.csv'
            posted = {'stream': (fname, open(self.datapath(fname))),
                      'encoding': u'utf-8',
                      'language_code': u'fr',
                      'delimiter': u'tab',
                      'format': u'lcsv'}
            req.form = self.fake_form('skos.scheme.import', posted, [(scheme, {})])
            # now actually tests the import, using scheme.view and not self.view which doesn't like
            # exception, even Redirect
            self.assertRaises(Redirect, scheme.view, 'skos.scheme.import')
        # check that the concept were added
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.find('ConceptScheme', title=u'musique').one()
            self.assertEqual(len(scheme.top_concepts), 2)
            concepts = cnx.find('Concept')
            self.assertEqual(len(concepts), 5)
            concept1 = cnx.find('Concept',
                                definition="Définition de l'organisation politique de l'organisme,"
                                ).one()
            label = concept1.preferred_label[0]
            self.assertEqual(len(concept1.preferred_label), 1)
            self.assertEqual((label.label, label.language_code), ('Vie politique', 'fr'))
            self.assertEqual(len(concept1.narrower_concept), 2)
            concept2 = cnx.find('Concept',
                                definition="Création volontaire ou en application de la loi").one()
            self.assertEqual(concept2.broader_concept[0], concept1)

    def _test_skos_view(self, vid, nb_expected_triples):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.find('ConceptScheme', title=u'musique').one()
            scheme.add_concept(u'pop')
            with NamedTemporaryFile(delete=False) as fobj:
                try:
                    fobj.write(scheme.view(vid))
                    fobj.close()
                    graph = default_graph()
                    graph.load('file://' + fobj.name, rdf_format='xml')
                finally:
                    os.unlink(fobj.name)
            triples = set(graph.triples())
            self.assertEqual(len(triples), nb_expected_triples, triples)

    def test_skos_primary_view(self):
        self._test_skos_view('primary.rdf', 5)

    def test_skos_list_view(self):
        self._test_skos_view('list.rdf', 2)


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos site customizations"""

from copy import copy
from datetime import datetime
import inspect

from logilab.common.decorators import monkeypatch

from cubicweb.dataimport import MetaGenerator, NoHookRQLObjectStore


@monkeypatch(MetaGenerator)
def __init__(self, cnx, baseurl=None, source=None):
    self._cnx = cnx
    if baseurl is None:
        config = cnx.vreg.config
        baseurl = config['base-url'] or config.default_base_url()
    if not baseurl[-1] == '/':
        baseurl += '/'
    self.baseurl = baseurl
    if source is None:
        source = cnx.repo.system_source
    self.source = source
    self.create_eid = cnx.repo.system_source.create_eid
    self.time = datetime.now()
    # attributes/relations shared by all entities of the same type
    self.etype_attrs = []
    self.etype_rels = []
    # attributes/relations specific to each entity
    self.entity_attrs = ['cwuri']
    schema = cnx.vreg.schema
    rschema = schema.rschema
    for rtype in self.META_RELATIONS:
        # skip owned_by / created_by if user is the internal manager
        if cnx.user.eid == -1 and rtype in ('owned_by', 'created_by'):
            continue
        if rschema(rtype).final:
            self.etype_attrs.append(rtype)
        else:
            self.etype_rels.append(rtype)


@monkeypatch(MetaGenerator)
def init_entity(self, entity):
    entity.eid = self.create_eid(self._cnx)
    extid = entity.cw_edited.get('cwuri')
    for attr in self.entity_attrs:
        if attr in entity.cw_edited:
            # already set, skip this attribute
            continue
        genfunc = self.generate(attr)
        if genfunc:
            entity.cw_edited.edited_attribute(attr, genfunc(entity))
    if isinstance(extid, unicode):
        extid = extid.encode('utf-8')
    return self.source, extid


@monkeypatch(NoHookRQLObjectStore)
def create_entity(self, etype, **kwargs):
    for k, v in kwargs.iteritems():
        kwargs[k] = getattr(v, 'eid', v)
    entity, rels = self.metagen.base_etype_dicts(etype)
    # make a copy to keep cached entity pristine
    entity = copy(entity)
    entity.cw_edited = copy(entity.cw_edited)
    entity.cw_clear_relation_cache()
    entity.cw_edited.update(kwargs, skipsec=False)
    entity_source, extid = self.metagen.init_entity(entity)
    cnx = self._cnx
    self.source.add_info(cnx, entity, entity_source, extid)
    self.source.add_entity(cnx, entity)
    kwargs = dict()
    if inspect.getargspec(self.add_relation).keywords:
        kwargs['subjtype'] = entity.cw_etype
    for rtype, targeteids in rels.iteritems():
        # targeteids may be a single eid or a list of eids
        inlined = self.rschema(rtype).inlined
        try:
            for targeteid in targeteids:
                self.add_relation(cnx, entity.eid, rtype, targeteid,
                                  inlined, **kwargs)
        except TypeError:
            self.add_relation(cnx, entity.eid, rtype, targeteids,
                              inlined, **kwargs)
    self._nb_inserted_entities += 1
    return entity

try:
    from cubicweb.devtools.testlib import CubicWebTC
except ImportError:
    # devtools not available
    pass
else:
    # #4985805
    from cubicweb.web import eid_param

    @monkeypatch(CubicWebTC, methodname='fake_form')
    @staticmethod
    def fake_form(formid, field_dict=None, entity_field_dicts=()):
        """Build _cw.form dictionnary to fake posting of some standard cubicweb form

        * `formid`, the form id, usually form's __regid__

        * `field_dict`, dictionary of name:value for fields that are not tied to an entity

        * `entity_field_dicts`, list of (entity, dictionary) where dictionary contains name:value
          for fields that are not tied to the given entity
        """
        assert field_dict or entity_field_dicts, \
            'field_dict and entity_field_dicts arguments must not be both unspecified'
        if field_dict is None:
            field_dict = {}
        form = {'__form_id': formid}
        fields = []
        for field, value in field_dict.items():
            fields.append(field)
            form[field] = value

        def _add_entity_field(entity, field, value):
            entity_fields.append(field)
            form[eid_param(field, entity.eid)] = value

        for entity, field_dict in entity_field_dicts:
            if '__maineid' not in form:
                form['__maineid'] = entity.eid
            entity_fields = []
            form.setdefault('eid', []).append(entity.eid)
            _add_entity_field(entity, '__type', entity.cw_etype)
            for field, value in field_dict.items():
                _add_entity_field(entity, field, value)
            if entity_fields:
                form[eid_param('_cw_entity_fields', entity.eid)] = ','.join(entity_fields)
        if fields:
            form['_cw_fields'] = ','.join(fields)
        return form



"""cubicweb-skos application package

"SKOS implementation for cubicweb"
"""
import sys


def register_skos_concept_rdf_list_mapping(reg):
    """Register minimal SKOS mapping at the concept level (for use in
    'list.rdf' views).
    """
    reg.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
    reg.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    reg.register_etype_equivalence('Concept', 'skos:Concept')
    reg.register_attribute_equivalence(
        'Concept', 'definition', 'skos:definition')
    reg.register_relation_equivalence(
        'Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')


def register_skos_concept_rdf_output_mapping(reg, with_labels=True):
    """Register full SKOS mapping at the concept level (for use in
    'primary.rdf' views).
    """
    register_skos_concept_rdf_list_mapping(reg)
    reg.register_attribute_equivalence('Concept', 'example', 'skos:example')
    reg.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:broader')
    reg.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:narrower',
                                      reverse=True)
    reg.register_relation_equivalence('Concept', 'related_concept', 'Concept', 'skos:related')
    # may be a concept instead of an ExternalUri if the concept is locally known, but that is
    # handled by the importer
    reg.register_relation_equivalence('Concept', 'exact_match', 'ExternalUri', 'skos:exactMatch')
    reg.register_relation_equivalence('Concept', 'close_match', 'ExternalUri', 'skos:closeMatch')


def register_skos_rdf_list_mapping(reg):
    """Register minimal SKOS mapping, only describing concept scheme (for use in 'list.rdf' views).
    """
    reg.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
    reg.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    reg.register_etype_equivalence('ConceptScheme', 'skos:ConceptScheme')
    reg.register_attribute_equivalence('ConceptScheme', 'title', 'dc:title')
    reg.register_attribute_equivalence('ConceptScheme', 'description', 'dc:description')


def register_skos_rdf_output_mapping(reg):
    """Register full SKOS mapping for outputing RDF from entities (for use in 'primary.rdf' views).

    Mapping doesn't include description for labels as it has to be handled manually, since yams
    models them as entities while SKOS uses literals.
    """
    register_skos_rdf_list_mapping(reg)
    register_skos_concept_rdf_output_mapping(reg, with_labels=False)


LABELS_RDF_MAPPING = {
    u'preferred_label': 'skos:prefLabel',
    u'alternative_label': 'skos:altLabel',
    u'hidden_label': 'skos:hiddenLabel',
}


def register_skos_rdf_input_mapping(reg):
    """Register SKOS mapping for input data.

    Mapping includes description for labels which should be included in `ExtEntity` for later
    transformation.
    """
    register_skos_rdf_list_mapping(reg)
    register_skos_rdf_output_mapping(reg)
    for rtype, predicate in LABELS_RDF_MAPPING.items():
        reg.register_relation_equivalence('Concept', rtype, 'Label', predicate)


def to_unicode(obj):
    """Turn some object (usually an exception) to unicode"""
    try:
        # The exception message might already be a unicode string.
        return unicode(obj)
    except UnicodeDecodeError:
        return str(obj).decode(sys.getdefaultencoding(), 'ignore')


class ExtEntity(object):
    """Transitional representation of an entity for use in data importer"""

    def __init__(self, etype, extid, values=None):
        self.etype = etype
        self.extid = extid
        if values is None:
            values = {}
        self.values = values
        self._schema = None

    def __repr__(self):
        return '<%s %s %s>' % (self.etype, self.extid, self.values)

    def prepare(self, schema):
        """Prepare an external entity for later insertion:

        * ensure attributes and inlined relations have a single value
        * turn set([value]) into value and remove key associated to empty set
        * remove non inlined relations and return them as a [(e1key, relation, e2key)] list

        Return a list of non inlined relations that may be inserted later, each relations defined by
        a 3-tuple (subject extid, relation type, object extid).

        Take care the importer may call this method several times.
        """
        if self._schema is not None:  # already prepared
            assert self._schema is schema
            return ()
        self._schema = schema
        eschema = schema.eschema(self.etype)
        deferred = []
        entity_dict = self.values
        for rtype in list(entity_dict):
            rschema = schema.rschema(rtype)
            if rschema.final or rschema.inlined:
                assert len(entity_dict[rtype]) <= 1, \
                    "more than one value for attribute %s: %s (%s)" % (rtype, entity_dict[rtype],
                                                                       self.extid)
                if entity_dict[rtype]:
                    entity_dict[rtype] = entity_dict[rtype].pop()
                    if (rschema.final and eschema.has_metadata(rtype, 'format')
                            and not rtype + '_format' in entity_dict):
                        entity_dict[rtype + '_format'] = u'text/plain'
                else:
                    del entity_dict[rtype]
            else:
                for target_extid in entity_dict[rtype]:
                    deferred.append((self.extid, rtype, target_extid))
                del entity_dict[rtype]
        return deferred

    def is_ready(self, extid2eid):
        """Return True if the ext entity is ready, i.e. has all the URIs used in inlined relations
        currently existing.
        """
        entity_dict = self.values
        assert self._schema, 'prepare() method should be called first'
        schema = self._schema
        for rtype in entity_dict:
            rschema = schema.rschema(rtype)
            if not rschema.final:
                # _prepare_extentity should drop other cases from the entity dict
                assert rschema.inlined
                if not entity_dict[rtype] in extid2eid:
                    return False
        # entity is ready, replace all relation's extid by eids
        for rtype in entity_dict:
            rschema = schema.rschema(rtype)
            if rschema.inlined:
                entity_dict[rtype] = extid2eid[entity_dict[rtype]]
        return True



# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from functools import wraps
from StringIO import StringIO

from logilab.common.testlib import require_module

from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos.entities import CSVIndentationError, CSVDecodeError
from cubes.skos.rdfio import (TYPE_PREDICATE_URI, RDFLibRDFGraph, LibRDFRDFGraph,
                              unicode_with_language as ul)


def get_narrower_concepts(concept):
    """Get a dictionnary matching the name of the concept to the dictionnary of its narrower
    concepts (recursive)
    """
    narrower_concepts = {}
    for concept in concept.narrower_concept:
        narrower_concepts[concept.dc_title()] = get_narrower_concepts(concept)
    return narrower_concepts


def build_result_dict(scheme):
    """Create a hierarchy with the concept names in a dictionary"""
    concepts = scheme.top_concepts
    result = {}
    for concept in concepts:
        result[concept.dc_title()] = get_narrower_concepts(concept)
    return result


class ConceptSchemeTC(CubicWebTC):

    def test_top_concepts(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'some classification')
            c1 = cnx.create_entity('Concept', in_scheme=scheme)
            cnx.create_entity('Label', label=u'hip', language_code=u'fr',
                              kind=u'preferred', label_of=c1)
            c2 = cnx.create_entity('Concept', in_scheme=scheme, broader_concept=c1)
            cnx.create_entity('Label', label=u'hop', language_code=u'fr',
                              kind=u'preferred', label_of=c2)
            cnx.commit()
            self.assertEqual(set(x.eid for x in scheme.top_concepts),
                             set((c1.eid,)))

    def test_add_concepts_from_file_interdoc(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'interdoc')
            cnx.commit()
        csv_file = self.datapath('thesaurus_interdoc_7_hierageneTab_shortened.csv')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            with open(csv_file) as sourcefile:
                scheme.add_concepts_from_file(sourcefile, u'utf-8', u'fr', u',')
            result = build_result_dict(scheme)
        expected = {
            u'ADMINISTRATION': {
                u'ACTION PUBLIQUE': {
                    u'ACTE ADMINISTRATIF': {
                        u'ACTE CREATEUR DE DROITS': {},
                        u'ACTE UNILATERAL': {
                            u'VISA': {},
                        },
                        u'CONTRAT PUBLIC': {
                            u'CONTRAT ADMINISTRATIF': {
                                u'CLAUSE EXORBITANTE': {},
                                u'COTRAITANCE': {
                                    u'GROUPEMENT D\'ENTREPRISES': {
                                        u'GROUPEMENT CONJOINT': {},
                                        u'GROUPEMENT SOLIDAIRE': {},
                                    },
                                },
                                u'EQUILIBRE FINANCIER': {
                                    u'FAIT DU PRINCE': {}
                                },
                                u'VENTE EN L\'ETAT FUTUR D\'ACHEVEMENT': {},
                            },
                            u'CONVENTION D\'EXPLOITATION': {},
                            u'PARTENARIAT PUBLIC-PRIVE': {
                                u'SOCIETE LOCALE DE PARTENARIAT': {},
                            },
                        },
                    },
                },
            },
            u'AMENAGEMENT': {
                u'AMENAGEMENT DU TERRITOIRE': {
                    u'AMENAGEMENT DE LA MONTAGNE': {
                        u'PROTECTION DE LA MONTAGNE': {
                            u'RTM': {},
                        },
                        u'UTN': {},
                        u'ZONE MONTAGNE': {},
                    },
                    u'AMENAGEMENT DU LITTORAL': {
                        u'BANDE LITTORALE DES 100 METRES': {},
                    },
                },
            },
        }
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_ok(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        csv_file = self.datapath('hierarchical_csv_example_shortened.csv')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            with open(csv_file) as sourcefile:
                scheme.add_concepts_from_file(sourcefile, u'utf-8', u'fr', u'\t')
            result = build_result_dict(scheme)
        expected = {u'APPELLATION DES LOIS ET DES RAPPORTS':
                    {u'LOIS RAPPORTS JURISPRUDENCE':
                     {u'APPELLATION DE DECISIONS DE JURISPRUDENCE':
                      {u'ARRET BERKANI': {}, u'ARRET TERNON': {}},
                      u'APPELLATION DES RAPPORTS':
                      {u'RAPPORT ARTHUIS': {}, u'LOLF': {}, u'LOV': {}}
                     }
                    },
                    u'LISTE DES MOTS OUTILS':
                    {u'CHAPITRE MOTS OUTILS':
                     {u'MOTS OUTILS':
                      {u'ABATTEMENT': {}, u'ACCORD': {}}
                     }
                    }
                   }
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_sep_inside_concept(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n\tti\tti\n\ttata')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u'\t')
            result = build_result_dict(scheme)
        expected = {u'toto': {u'ti\tti': {},
                              u'tata' : {}}}
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_decode_error(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        with open(self.datapath('bad_encoding.csv')) as fobj:
            with self.admin_access.web_request() as req:
                scheme = req.entity_from_eid(scheme.eid)
                with self.assertRaises(CSVDecodeError) as cm:
                    scheme.add_concepts_from_file(fobj, u'utf-8', u'fr', u',')
                self.assertEqual(cm.exception.line, 3)
                # Now with the good encoding.
                fobj.seek(0)
                scheme.add_concepts_from_file(fobj, u'latin1', u'fr', u',')
                result = build_result_dict(scheme)
        expected = {u'voici mes concepts en latin1': {u'celui-ci est bon': {},
                                                      u'celui-là n\'est pas bon': {}}}
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_wrong_indentation(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n\ttiti\n\t\t\ttata')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            with self.assertRaises(CSVIndentationError) as cm:
                scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u'\t')
            self.assertEqual(cm.exception.line, 3)

    def test_add_concepts_from_file_multiple_deindentation(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n\ttiti\n\t\ttata\ntutu')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u'\t')
            result = build_result_dict(scheme)
        expected = {u'toto': {u'titi': {u'tata': {}}},
                    u'tutu': {}}
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_sep_comma(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n,titi\n,,tata\n,,  tati  \n tutu \n,tuti\n,,titu\n,toti\n')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u',')
            result = build_result_dict(scheme)
        expected = {u'toto': {u'titi': {u'tata': {}, u'tati': {}}},
                    u'tutu': {u'tuti': {u'titu': {}}, u'toti': {}}}
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_sep_space(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n titi   iti\n  tata\n  tati  \ntutu \n tuti\n  titu\n toti\n')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u' ')
            result = build_result_dict(scheme)
        expected = {u'toto': {u'titi   iti': {u'tata': {}, u'tati': {}}},
                    u'tutu': {u'tuti': {u'titu': {}}, u'toti': {}}}
        self.assertEqual(result, expected)


class ConceptTC(CubicWebTC):
    def setUp(self):
        super(ConceptTC, self).setUp()
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'some classification')
            c1 = cnx.create_entity('Concept', in_scheme=scheme)
            cnx.create_entity('Label', label=u'hîp', language_code=u'fr-CA',
                              kind=u'preferred', label_of=c1)
            cnx.create_entity('Label', label=u'hip', language_code=u'en',
                              kind=u'preferred', label_of=c1)
            cnx.create_entity('Label', label=u'hop', language_code=u'fr',
                              kind=u'alternative', label_of=c1)
            cnx.commit()
        self.c1_eid = c1.eid

    def test_dc_title(self):
        with self.admin_access.client_cnx() as cnx:
            c1 = cnx.entity_from_eid(self.c1_eid)
            self.assertEqual(c1.dc_title(), u'hip')
            cnx.lang = 'fr'
            self.assertEqual(c1.dc_title(), u'hîp')

    def test_labels(self):
        with self.admin_access.client_cnx() as cnx:
            c1 = cnx.entity_from_eid(self.c1_eid)
            self.assertEqual(c1.labels,
                             {u'en': u'hip', u'fr': u'hîp'})


SKOS_SCHEME = 'http://www.w3.org/2004/02/skos/core#ConceptScheme'
SKOS_CONCEPT = 'http://www.w3.org/2004/02/skos/core#Concept'
SKOS_INSCHEME = 'http://www.w3.org/2004/02/skos/core#inScheme'
SKOS_LABEL = 'http://www.w3.org/2004/02/skos/core#prefLabel'
SKOS_DEFINITION = 'http://www.w3.org/2004/02/skos/core#definition'
DC_TITLE = 'http://purl.org/dc/elements/1.1/title'


def require_lib(lib):
    """Decorate a test method with require_module(lib) and bind it respective
    graph generator instance.
    """
    graphcls = {'rdflib': RDFLibRDFGraph,
                'RDF': LibRDFRDFGraph}[lib]
    def decorator(testfunc):
        @wraps(testfunc)
        def wrapper(self):
            return testfunc(self, graphcls())
        return require_module(lib)(wrapper)
    return decorator


class RDFAdaptersTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'some classification')
            c1 = cnx.create_entity('Concept', in_scheme=scheme,
                                   definition=u'def1')
            cnx.create_entity('Label', label=u'hip', language_code=u'fr',
                              kind=u'preferred', label_of=c1)
            c2 = cnx.create_entity('Concept', in_scheme=scheme, broader_concept=c1)
            cnx.create_entity('Label', label=u'hop', language_code=u'fr',
                              kind=u'preferred', label_of=c2)
            cnx.commit()
            self.assertEqual(set(x.eid for x in scheme.top_concepts),
                             set((c1.eid,)))

    def _test_concept_list_repr(self, graph):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.execute('ConceptScheme X').one()
            concept = cnx.find('Concept', definition=u'def1').one()
            rdfgenerator = concept.cw_adapt_to('RDFList')
            rdfgenerator.fill(graph)
            expected = [
                (concept.absolute_url(), TYPE_PREDICATE_URI, SKOS_CONCEPT),
                (concept.absolute_url(), SKOS_DEFINITION, u'def1'),
                (concept.absolute_url(), SKOS_INSCHEME, scheme.absolute_url()),
        ]
        self.assertCountEqual(set(graph.triples()), expected)

    test_rdflib_concept_list_repr = require_lib('rdflib')(_test_concept_list_repr)
    test_librdf_concept_list_repr = require_lib('RDF')(_test_concept_list_repr)

    def _test_scheme_list_repr(self, graph):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.execute('ConceptScheme X').one()
            rdfgenerator = scheme.cw_adapt_to('RDFList')
            rdfgenerator.fill(graph)
        expected = [(scheme.absolute_url(), TYPE_PREDICATE_URI, SKOS_SCHEME),
                    (scheme.absolute_url(), DC_TITLE, u'some classification')]
        self.assertCountEqual(graph.triples(), expected)

    test_rdflib_scheme_list_repr = require_lib('rdflib')(_test_scheme_list_repr)
    test_librdf_scheme_list_repr = require_lib('RDF')(_test_scheme_list_repr)

    def _test_scheme_primary_repr(self, graph):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.execute('ConceptScheme X').one()
            c1 = scheme.top_concepts[0]
            c2 = c1.narrower_concept[0]
            scheme_uri = scheme.absolute_url()
            c1_uri = c1.absolute_url()
            c2_uri = c2.absolute_url()
            rdfgenerator = scheme.cw_adapt_to('RDFPrimary')
            rdfgenerator.fill(graph)
        expected = [
            (scheme_uri, TYPE_PREDICATE_URI, SKOS_SCHEME),
            (scheme_uri, DC_TITLE, u'some classification'),
            (c1_uri, TYPE_PREDICATE_URI, SKOS_CONCEPT),
            (c1_uri, SKOS_INSCHEME, scheme_uri),
            (c1_uri, SKOS_LABEL, ul('hip', 'fr')),
            (c1_uri, SKOS_DEFINITION, u'def1'),
            (c1_uri, u'http://www.w3.org/2004/02/skos/core#narrower', c2_uri),
            (c2_uri, TYPE_PREDICATE_URI, SKOS_CONCEPT),
            (c2_uri, SKOS_INSCHEME, scheme_uri),
            (c2_uri, SKOS_LABEL, ul('hop', 'fr')),
            (c2_uri, u'http://www.w3.org/2004/02/skos/core#broader', c1_uri),
        ]
        self.assertCountEqual(graph.triples(), expected)

    test_rdflib_scheme_primary_repr = require_lib('rdflib')(_test_scheme_primary_repr)
    test_librdf_scheme_primary_repr = require_lib('RDF')(_test_scheme_primary_repr)

    def _test_concept_primary_repr(self, graph):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.execute('ConceptScheme X').one()
            c1 = cnx.find('Concept', definition=u'def1').one()
            c2 = c1.narrower_concept[0]
            c1_uri, c2_uri = c1.absolute_url(), c2.absolute_url()
            rdfgenerator = c1.cw_adapt_to('RDFPrimary')
            rdfgenerator.fill(graph)
        expected = [
            (c1_uri, TYPE_PREDICATE_URI, SKOS_CONCEPT),
            (c1_uri, SKOS_INSCHEME, scheme.absolute_url()),
            (c1_uri, SKOS_DEFINITION, u'def1'),
            (c1_uri, SKOS_LABEL, ul('hip', 'fr')),
            (c1_uri, u'http://www.w3.org/2004/02/skos/core#narrower', c2_uri),
            (c2_uri, u'http://www.w3.org/2004/02/skos/core#broader', c1_uri),
        ]
        self.assertCountEqual(set(graph.triples()), expected)

    test_rdflib_concept_primary_repr = require_lib('rdflib')(_test_concept_primary_repr)
    test_librdf_concept_primary_repr = require_lib('RDF')(_test_concept_primary_repr)

    def test_warm_relation_cache(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.execute('ConceptScheme X').one()
            rdfgenerator = scheme.cw_adapt_to('RDFPrimary')
            rdfgenerator.warm_caches()
            c1 = cnx.execute('Concept X WHERE X preferred_label L, L label "hip"').one()
            c2 = cnx.execute('Concept X WHERE X preferred_label L, L label "hop"').one()
            scheme._cw.execute = lambda rql, args=None: 1 / 0  # ensure no RQL query is done
            self.assertCountEqual(scheme.cw_attr_cache,
                                  ['cwuri', 'creation_date', 'modification_date',
                                   'title', 'description', 'description_format'])
            self.assertEqual(len(scheme.reverse_in_scheme), 2)
            self.assertEqual(c1.scheme.eid, scheme.eid)
            self.assertIn('cwuri', c1.cw_attr_cache)
            self.assertIn('definition', c1.cw_attr_cache)
            self.assertIn('example', c1.cw_attr_cache)
            self.assertEqual(c1.labels, {u'fr': u'hip'})
            for rtype in ('exact_match', 'close_match'):
                self.assertIn(rtype + '_subject', c1._cw_related_cache)
                self.assertNotIn(rtype + '_object', c1._cw_related_cache)
            self.assertEqual(len(c1.broader_concept), 0)
            self.assertEqual(len(c1.reverse_broader_concept), 1)
            self.assertEqual(len(c1.related_concept), 0)
            self.assertEqual(c1.reverse_broader_concept, (c2,))
            self.assertEqual(len(c2.broader_concept), 1)
            self.assertEqual(len(c2.reverse_broader_concept), 0)
            self.assertEqual(c2.broader_concept, (c1,))


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""SKOS import is available through:

* a c-c command
* a repository service
* the datafeed source API

The first two will create local entities (i.e. whose `cw_source` relation points to system source)
while the last will create entities belonging to the external source.

This file contains generic import code, the datafeed parser and the repository service.
"""

from itertools import imap
from hashlib import md5
from uuid import uuid4

from cubicweb import schema
from cubicweb.predicates import match_kwargs, match_user_groups
from cubicweb.server import Service
from cubicweb.server.sources import datafeed

from cubicweb import dataimport
from cubes.skos import LABELS_RDF_MAPPING, ExtEntity, register_skos_rdf_input_mapping, rdfio, lcsv
from cubes.skos.dataimport import ExtEntitiesImporter, HTMLImportLog, cwuri2eid


def dump_relations(cnx, eid, etype):
    """Return a list of relation 3-uples `(subject_eid, relation, object_eid)` with None instead of
    `subject_eid` or `object_eid` depending on whether the entity type corresponding to `eid` is
    subject or object.
    """
    eschema = cnx.vreg.schema.eschema(etype)
    relations = []
    for rschema, _, role in eschema.relation_definitions():
        if rschema.rule:  # computed relation
            continue
        rtype = rschema.type
        if rtype in schema.VIRTUAL_RTYPES or rtype in schema.META_RTYPES:
            continue
        if role == 'subject':
            for object_eid, in cnx.execute('Any Y WHERE X %s Y, X eid %%(x)s' % rtype, {'x': eid}):
                if object_eid == eid:
                    object_eid = None
                relations.append((None, rtype, object_eid))
        else:
            for subject_eid, in cnx.execute('Any Y WHERE Y %s X, X eid %%(x)s' % rtype, {'x': eid}):
                if subject_eid == eid:
                    subject_eid = None
                relations.append((subject_eid, rtype, None))
    return relations


class RDFSKOSImportService(Service):
    """Cubicweb service to import a ConceptScheme from SKOS RDF"""
    __regid__ = 'rdf.skos.import'
    __select__ = match_kwargs('stream') & match_user_groups('managers')

    def call(self, stream, **kwargs):
        import_log = HTMLImportLog(getattr(stream, 'filename', u''))
        (created, updated), scheme_uris = self._do_import(stream, import_log, **kwargs)
        import_log.record_info('%s entities created' % len(created))
        import_log.record_info('%s entities updated' % len(updated))
        self._cw.commit()
        return import_log.logs, scheme_uris

    def _do_import(self, stream, import_log, **kwargs):
        """Extracted method to let a chance to client cubes to give more arguments to _skos_import
        function
        """
        return _skos_import_rdf(self._cw, stream, import_log, **kwargs)


class LCSVSKOSImportService(RDFSKOSImportService):
    """Cubicweb service to import a Concept and Label entities from LCSV file"""
    __regid__ = 'lcsv.skos.import'

    def _do_import(self, stream, import_log, **kwargs):
        """Import Concept and Label entities and related them to `scheme`"""
        return _skos_import_lcsv(self._cw, stream, import_log, **kwargs)


class SKOSParser(datafeed.DataFeedParser):
    __regid__ = 'rdf.skos'

    def process(self, url, raise_on_error=False):
        """Main entry point"""
        try:
            created, updated = self._do_import(url, raise_on_error=raise_on_error)[0]
        except Exception as ex:
            if raise_on_error:
                raise
            self.exception('error while importing %s', url)
            self.import_log.record_error(str(ex))
            return True
        self.stats['created'] = created
        self.stats['updated'] = updated
        return False

    def _do_import(self, url, raise_on_error, **kwargs):
        """Extracted method to let a chance to client cubes to give more arguments to _skos_import
        function
        """
        try:
            rdf_format = rdfio.guess_rdf_format(url)
        except ValueError:
            rdf_format = 'xml'
        return _skos_import_rdf(self._cw, url, self.import_log, source=self.source,
                                rdf_format=rdf_format, raise_on_error=raise_on_error, **kwargs)


def _skos_import_lcsv(cnx, stream, import_log, scheme_uri, delimiter,
                      encoding, language_code=None, **kwargs):
    """import SKOS from LCSV stream or URL (by transforming it to RDF first)"""
    graph = rdfio.RDFLibRDFGraph()
    # add LCSV statements to the RDF graph
    lcsv2rdf = lcsv.LCSV2RDF(stream, delimiter, encoding, default_lang=language_code,
                             uri_cls=graph.uri, uri_generator=lambda x: str(uuid4()) + x)
    for (subj, pred, obj) in lcsv2rdf.triples():
        graph.add(subj, pred, obj)

    # we need an extra transform that link Concept to the parent concept skeme
    def relate_concepts(extentity):
        """Relate Concept ExtEntities to the ConceptScheme"""
        if extentity.etype == 'Concept':
            extentity.values.setdefault('in_scheme', set([])).add(scheme_uri)
        return extentity

    entities = imap(relate_concepts, _skos_entities(graph))
    # now import the extentities stream
    return _skos_import(cnx, entities, import_log, use_extid_as_cwuri=False, **kwargs)


def _skos_import_rdf(cnx, stream_or_url, import_log, rdf_format=None, **kwargs):
    """import SKOS from RDF stream or URL"""
    # build the RDF graph
    graph = rdfio.RDFLibRDFGraph()
    graph.load(stream_or_url, rdf_format)
    entities = _skos_entities(graph)
    return _skos_import(cnx, entities, import_log, **kwargs)


def _skos_entities(graph):
    """return external entities from a SKOS stream or URL"""
    reg = rdfio.RDFRegistry()
    register_skos_rdf_input_mapping(reg)
    all_label_rtypes = frozenset(LABELS_RDF_MAPPING)
    # transform direct mapping of RDF as ExtEntity into Yams compatible ExtEntity by turning label
    # literals into entities.
    for extentity in rdfio.rdf_graph_to_entities(reg, graph, ('ConceptScheme', 'Concept')):
        label_rtypes = frozenset(extentity.values) & all_label_rtypes
        if not label_rtypes:
            yield extentity
            continue
        labels = []
        for rtype in label_rtypes:
            kind = rtype.split('_', 1)[0]  # drop '_label' suffix
            for label in extentity.values.pop(rtype):
                lang = getattr(label, 'lang', None)
                md5hash = md5(str(lang) + label.encode('utf-8'))
                labelid = str(extentity.extid) + '#' + rtype + md5hash.hexdigest()
                labels.append(ExtEntity('Label', labelid,
                                        {'label': set([label]),
                                         'language_code': set([lang]),
                                         'kind': set([kind]),
                                         'label_of': set([extentity.extid])}))
        # yield extentity before labels since it must be handled first in case the ExternalUri
        # to Concept transformation apply
        yield extentity
        for label in labels:
            yield label


def _skos_import(cnx, entities, import_log, source=None, raise_on_error=False, metagenerator=None,
                 use_extid_as_cwuri=True):
    """import SKOS external entities"""
    # use NoHookRQLObjectStore for insertion in the database, with a custom metagenerator to
    # properly set the source for created entities
    if metagenerator is None:
        metagenerator = dataimport.MetaGenerator(cnx, source=source)
    store = dataimport.NoHookRQLObjectStore(cnx, metagenerator)
    # only consider the system source for schemes and labels
    if source is None:
        source_eid = cnx.repo.system_source.eid
    else:
        source_eid = source.eid
    extid2eid = cwuri2eid(cnx, ('ConceptScheme', 'Label'), source_eid=source_eid)
    # though concepts and external URIs may come from any source
    extid2eid.update(cwuri2eid(cnx, ('Concept', 'ExternalUri')))
    # plug function that turn previously known external uris by newly inserted concepts
    restore_relations = {}

    def externaluri_to_concept(extentity, cnx=cnx, extid2eid=extid2eid,
                               restore_relations=restore_relations):
        try:
            eid = extid2eid[extentity.extid]
        except KeyError:
            pass
        else:
            if extentity.etype == 'Concept' and cnx.entity_metas(eid)['type'] == 'ExternalUri':
                # We have replaced the external uri by the new concept. As entities.extid column is
                # unique, we've to drop the external uri before inserting the concept, so we:
                #  1. record every relations from/to the external uri,
                #  2. remove it,
                #  3. insert the concept and
                #  4. reinsert relations using the concept instead
                #
                # 1. record relations from/to the external uri
                restore_relations[extentity.extid] = dump_relations(cnx, eid, 'ExternalUri')
                # 2. remove the external uri entity
                cnx.execute('DELETE ExternalUri X WHERE X eid %(x)s', {'x': eid})
                # 3. drop its extid from the mapping to trigger insertion of the concept by the
                # importer
                del extid2eid[extentity.extid]
                # 4. will be done in SKOSExtEntitiesImporter
        return extentity

    entities = imap(externaluri_to_concept, entities)
    # plug function to detect the concept scheme
    concept_schemes = []

    def record_scheme(extentity):
        if extentity.etype == 'ConceptScheme':
            concept_schemes.append(extentity.extid)
        return extentity

    entities = imap(record_scheme, entities)
    etypes_order_hint = ('ConceptScheme', 'Concept', 'Label')
    importer = SKOSExtEntitiesImporter(cnx, store, import_log, source=source, extid2eid=extid2eid,
                                       raise_on_error=raise_on_error,
                                       etypes_order_hint=etypes_order_hint,
                                       restore_relations=restore_relations)
    stats = importer.import_entities(entities, use_extid_as_cwuri=use_extid_as_cwuri)
    return stats, concept_schemes


class SKOSExtEntitiesImporter(ExtEntitiesImporter):
    """Override ExtEntitiesImporter to handle creation of additionnal relations to newly created
    concepts that replace a former external uri, and to create ExternalUri entities for URIs used in
    exact_match / close_match relations which have no known entity in the repository yet.
    """

    def __init__(self, *args, **kwargs):
        self.restore_relations = kwargs.pop('restore_relations')
        super(SKOSExtEntitiesImporter, self).__init__(*args, **kwargs)

    def create_entity(self, extentity):
        entity = super(SKOSExtEntitiesImporter, self).create_entity(extentity)
        # (4.) restore relations formerly from/to an equivalent external uri
        try:
            relations = self.restore_relations.pop(extentity.extid)
        except KeyError:
            return entity
        for subject_eid, rtype, object_eid in relations:
            if subject_eid is None:
                subject_eid = entity.eid
            if object_eid is None:
                object_eid = entity.eid
            self.store.relate(subject_eid, rtype, object_eid)
        return entity

    def create_deferred_relations(self, deferred):
        # create missing targets for exact_match and close_match relations
        for rtype in ('exact_match', 'close_match'):
            relations = deferred.get(rtype, ())
            for subject_uri, object_uri in relations:
                if object_uri not in self.extid2eid:
                    extentity = ExtEntity('ExternalUri', object_uri,
                                          values=dict(uri=unicode(object_uri),
                                                      cwuri=unicode(object_uri)))
                    assert self.create_entity(extentity).cw_etype == 'ExternalUri'
        return super(SKOSExtEntitiesImporter, self).create_deferred_relations(deferred)

    def existing_relations(self, rtype):
        """return a set of (subject, object) eids already related by `rtype`"""
        if rtype in ('exact_match', 'close_match'):
            rql = 'Any X,O WHERE X %s O' % rtype
            return set(tuple(x) for x in self.cnx.execute(rql))
        return super(SKOSExtEntitiesImporter, self).existing_relations(rtype)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos unit tests for hooks"""

from yams import ValidationError
from cubicweb.devtools.testlib import CubicWebTC


class HooksTC(CubicWebTC):

    def test_externaluri_to_concept(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'preexisting')
            concept = scheme.add_concept(u'something',
                                         cwuri=u'http://data.bnf.fr/ark:/12148/cb11934798x')
            exturi = cnx.create_entity('ExternalUri', cwuri=u'http://someuri/someobject',
                                       uri=u'http://someuri/someobject',
                                       exact_match=concept)
            cnx.commit()
            # now insert a concept with the external uri as cwuri
            scheme = cnx.create_entity('ConceptScheme')
            new_concept = scheme.add_concept(u'some object', cwuri=u'http://someuri/someobject')
            cnx.commit()
            # ensure the external uri has been replaced by the concept and deleted
            concept.cw_clear_all_caches()
            self.assertEqual(concept.exact_match[0].eid, new_concept.eid)
            self.failIf(cnx.execute('Any X WHERE X eid %(x)s', {'x': exturi.eid}))

    def test_skos_source(self):
        with self.admin_access.repo_cnx() as cnx:
            skos = cnx.create_entity('SKOSSource', name=u'a source',
                                     url=u'http://skos.org\nhttp://skos.net')
            cnx.commit()
            source = cnx.find('CWSource', name=u'a source').one()
            self.assertEqual(source.url, u'http://skos.org\nhttp://skos.net')
            skos.cw_set(name=u'new name', url=u'http://skos.com')
            self.assertEqual(source.name, u'new name')
            self.assertEqual(source.url, u'http://skos.com')

    def check_duplicated_preferred_label(self, language_code, expected_errors):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'scheme')
            concept = scheme.add_concept(u'un concept', language_code=language_code)
            cnx.create_entity('Label', label=u'le concept', kind=u'preferred', label_of=concept,
                              language_code=language_code)
            with self.assertRaises(ValidationError) as cm:
                cnx.commit()
        self.assertEqual(cm.exception.errors, expected_errors)

    def test_concept_unique_pref_label_language(self):
        self.check_duplicated_preferred_label(
            u'fr', {'': u'a preferred label in "%(lang)s" language already exists',
                    'language_code': u'please use another language code'})

    def test_concept_unique_pref_label_no_language(self):
        self.check_duplicated_preferred_label(
            None, {'': u'a preferred label without language already exists',
                   'language_code': u'please specify a language code'})



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos schema"""

from yams.buildobjs import (EntityType, RelationDefinition, ComputedRelation,
                            String, RichString)

from cubicweb.schema import RQLConstraint, RRQLExpression

_ = unicode


class ConceptScheme(EntityType):
    """Vocabulary made of concepts"""
    title = String(unique=True, fulltextindexed=True)
    description = RichString(fulltextindexed=True)


class Concept(EntityType):
    """Unit of thought: idea, meaning, or (category of) object(s) and event(s)"""
    # URI of the concept is stored in the standard cwuri attribute
    definition = RichString(fulltextindexed=True)
    example = RichString(fulltextindexed=True)


class Label(EntityType):
    """Expression used to refer to a concept in natural language"""
    label = String(required=True, fulltextindexed=True)
    kind = String(required=True, internationalizable=True,
                  vocabulary=[_('preferred'), _('alternative'), _('hidden')],
                  default=u'preferred')
    language_code = String(maxsize=5,
                           description=_('language code, e.g. "en" or "fr-fr"'))


class in_scheme(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'delete': (RRQLExpression('O cw_source SO, SO name = "system"'), ),
        'add': (RRQLExpression('O cw_source SO, SO name = "system"'), ),
    }
    subject = 'Concept'
    object = 'ConceptScheme'
    cardinality = '+*'
    composite = 'object'
    description = _('scheme the concept has been attached to')


class label_of(RelationDefinition):
    subject = 'Label'
    object = 'Concept'
    cardinality = '1+'
    inlined = True
    composite = 'object'
    description = _('concept label')
    fulltext_container = 'object'


class preferred_label(ComputedRelation):
    rule = 'O label_of S, O kind "preferred"'


class alternative_label(ComputedRelation):
    rule = 'O label_of S, O kind "alternative"'


class hidden_label(ComputedRelation):
    rule = 'O label_of S, O kind "hidden"'


class broader_concept(RelationDefinition):
    subject = 'Concept'
    object = 'Concept'
    cardinality = '**'
    description = _('more general concept')
    constraints = [RQLConstraint('S in_scheme CS, O in_scheme CS')]


class narrower_concept(ComputedRelation):
    rule = 'O broader_concept S'


class related_concept(RelationDefinition):
    subject = 'Concept'
    object = 'Concept'
    description = _('related concept')
    symmetric = True
    constraints = [RQLConstraint('S in_scheme CS, O in_scheme CS')]


class exact_match(RelationDefinition):
    subject = 'Concept'
    object = ('Concept', 'ExternalUri')
    description = _('equivalent concept in another vocabulary')
    symmetric = True


class close_match(RelationDefinition):
    subject = 'Concept'
    object = ('Concept', 'ExternalUri')
    description = _('close concept in another vocabulary')


class SKOSSource(EntityType):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', ),
        'update': ('managers', ),
        'delete': ('managers', ),
    }
    # description are the same as for CWSource attributes hence shouldn't be marked for translation
    # here
    name = String(required=True, unique=True, maxsize=128,
                  description=_('name of the source'))
    url = String(description=_('URLs from which content will be imported. '
                               'You can put one url per line'))


class through_cw_source(RelationDefinition):
    __permissions__ = {'read': ('managers', 'users'),
                       'add': (),
                       'delete': ()}
    subject = 'SKOSSource'
    object = 'CWSource'
    cardinality = '1*'
    composite = 'subject'
    inlined = True



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos entity classes"""

from logilab.common.decorators import cachedproperty

from cubicweb.predicates import is_instance
from cubicweb.view import EntityAdapter
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.entities.adapters import ITreeAdapter

from cubes import skos
from cubes.skos import rdfio
from cubes.skos.rdfio import unicode_with_language as ul


def _add_concept(scheme, label, language_code, **kwargs):
    cnx = scheme._cw
    concept = cnx.create_entity('Concept', in_scheme=scheme, **kwargs)
    cnx.create_entity('Label', label=label, language_code=language_code,
                      kind=u'preferred', label_of=concept)
    # hooks do use existing labels, clear cache here to ensure the label is considered
    concept.cw_clear_all_caches()
    return concept


class CSVParseError(Exception):
    """Error during CSV parsing, with a line information."""

    def __init__(self, line):
        super(CSVParseError, self).__init__()
        self.line = line


class CSVIndentationError(CSVParseError):
    """Invalid indentation (level was reduced by more than one indentation)."""


class CSVDecodeError(CSVParseError):
    """Decode exception, probably due to a wrongly specified encoding."""


class ConceptScheme(AnyEntity):
    __regid__ = 'ConceptScheme'
    fetch_attrs, cw_fetch_order = fetch_config(('title', 'cwuri'))

    def dc_title(self):
        if self.title:
            return self.title
        return self.cwuri

    @cachedproperty
    def top_concepts_rset(self):
        return self._cw.execute(
            'Any C,CU WHERE C cwuri CU, C in_scheme X, NOT C broader_concept SC, X eid %(x)s',
            {'x': self.eid})

    @property
    def top_concepts(self):
        return list(self.top_concepts_rset.entities())

    def add_concept(self, label, language_code=u'en', **kwargs):
        """Add a top-concept to this scheme"""
        return _add_concept(self, label, language_code, **kwargs)

    def add_concepts_from_file(self, stream, encoding, language_code, delimiter):
        """Read a stream and create the listed concepts inside the ConceptScheme

        'delimiters' are considered as hierarchical information. There must be a concept per line,
        each line starting by N delimiters (indicating the hierarchical level) or nothing if the
        concept has no parent.

        Example (delimiter = u';')
        titi
        ;toto
        ;;tata
        --> ok: titi is a top-concept of the scheme, toto is a narrower concept of titi, tata is a
        narrower concept of toto

        titi
         ;toto
        --> 'titi' and ' ;toto' will be considered as two concepts of the scheme.
        """
        # This ordered list behaves like a state machine. It will contain all the concepts from
        # the conceptscheme to the more recent broader concept. The last element is consequently
        # the current parent concept. When de-indenting, the concepts are popped.
        level = -1
        broaderconcepts = [(self, level)]
        for nline, line in enumerate(stream, 1):
            try:
                line = line.rstrip().decode(encoding)
            except UnicodeDecodeError:
                raise CSVDecodeError(nline)
            if not line:
                continue
            # Find the label by removing leading delimiters (and spaces), then
            # remove trailing delimiters and spaces.
            label = line.lstrip(delimiter + ' ').rstrip(delimiter + ' ')
            if not label:
                # The line is full of delimiter.
                continue
            next_level = line[:line.index(label)].count(delimiter)
            if next_level - level > 1:
                # Concept must be at much one level below its parent.
                raise CSVIndentationError(nline)
            elif next_level <= level:
                # Walk back levels.
                while next_level <= level:
                    _, level = broaderconcepts.pop()
            concept = broaderconcepts[-1][0].add_concept(label, language_code)
            broaderconcepts.append((concept, level))
            level = next_level


class Concept(AnyEntity):
    __regid__ = 'Concept'
    fetch_attrs, cw_fetch_order = fetch_config(('cwuri',))

    def dc_title(self):
        return self.label()

    @property
    def parent_concept(self):
        return self.broader_concept[0] if self.broader_concept else None

    @property
    def scheme(self):
        return self.in_scheme[0]

    @property
    def labels(self):
        return dict((l.short_language_code, l.label) for l in self.preferred_label)

    def label(self, language_code=None, default_language_code='en'):
        if language_code is None:
            language_code = self._cw.lang
        try:
            return self.labels[language_code[:2].lower()]
        except KeyError:
            try:
                return self.labels[default_language_code[:2].lower()]
            except KeyError:
                # pick one random label
                return self.labels.values()[0]

    def add_concept(self, label, language_code=u'en'):
        """Add a sub-concept to this concept"""
        return _add_concept(self.scheme, label, language_code, broader_concept=self)


class Label(AnyEntity):
    __regid__ = 'Label'
    fetch_attrs, cw_fetch_order = fetch_config(('language_code', 'label'))

    @property
    def short_language_code(self):
        """Return the 2 letters language code for this label"""
        if self.language_code is None:
            return None
        return self.language_code[:2].lower()


class ConceptITreeAdapter(ITreeAdapter):
    """ITree adapater for Concept"""
    __select__ = is_instance('Concept')
    tree_relation = 'broader_concept'


class AbstractRDFAdapter(EntityAdapter):
    """Abstract RDF adapter, providing helper methods easing usage of :class:rdfio.RDFRegistry to
    generate (part of) the graph.
    """
    __abstract__ = True

    @cachedproperty
    def registry(self):
        """Return the RDFRegistry to be used for RDF serialisation."""
        reg = rdfio.RDFRegistry()
        self.register_rdf_mapping(reg)
        return reg

    def register_rdf_mapping(self, reg):
        """Add mapping RDF serialisation to the given registry."""
        raise NotImplementedError

    def graph_walker(self):
        """Return a walker that control the processing of the entities graph."""
        return rdfio.GraphWalker([self.entity], accept=self.accept)

    def accept(self, entity):
        """Return True if the entity should be recursivly added to the graph."""
        return False

    def add_entities_to_rdf_graph(self, graph):
        """Add entities mapped in the registry to the RDF graph."""
        walker = self.graph_walker()
        generator = rdfio.RDFGraphGenerator(graph, walker)
        generator.add_entities(self.registry)

    def fill(self, graph):
        """Adapter entry point, for use in RDF views."""
        self.add_entities_to_rdf_graph(graph)


class ConceptRDFListAdapter(AbstractRDFAdapter):
    """The RDFList adapter for Concept entity type."""

    __regid__ = 'RDFList'
    __select__ = is_instance('Concept')
    register_rdf_mapping = staticmethod(skos.register_skos_concept_rdf_list_mapping)


class ConceptRDFPrimaryAdapter(AbstractRDFAdapter):
    """The RDFList adapter for Concept entity type."""

    __regid__ = 'RDFPrimary'
    __select__ = is_instance('Concept')
    register_rdf_mapping = staticmethod(skos.register_skos_concept_rdf_output_mapping)

    def fill(self, graph):
        super(ConceptRDFPrimaryAdapter, self).fill(graph)
        reg = self.registry
        concept = self.entity
        concept_uri = graph.uri(concept.absolute_url())
        # manually handle labels to provide string with language code
        for rtype in skos.LABELS_RDF_MAPPING:
            skos_rel = graph.uri(reg.normalize_uri(skos.LABELS_RDF_MAPPING[rtype]))
            for label in getattr(concept, rtype):
                graph.add(concept_uri, skos_rel, ul(label.label, label.language_code))


class ConceptSchemeRDFListAdapter(AbstractRDFAdapter):
    """The RDFList adapter fill the RDF graph with a short representation of the adapted entity, for
    use with the 'list.rdf' view.
    """
    __regid__ = 'RDFList'
    __select__ = is_instance('ConceptScheme')

    register_rdf_mapping = staticmethod(skos.register_skos_rdf_list_mapping)


class ConceptSchemeRDFPrimaryAdapter(AbstractRDFAdapter):
    """The RDFPrimary adapter fill the RDF graph with a short representation of the adapted entity,
    for use with the 'primary.rdf' view.
    """
    __regid__ = 'RDFPrimary'
    __select__ = is_instance('ConceptScheme')

    register_rdf_mapping = staticmethod(skos.register_skos_rdf_output_mapping)

    def fill(self, graph):
        # prefill entity cache
        self.warm_caches()
        self.add_entities_to_rdf_graph(graph)
        for concept in self.entity.reverse_in_scheme:
            concept.cw_adapt_to('RDFPrimary').fill(graph)

    concept_attributes = 'cwuri definition example'.split()
    label_attributes = 'label language_code'.split()

    def warm_caches(self):
        scheme = self.entity
        cnx = self._cw
        scheme.complete()
        concept_rql = _select_attributes('Any CS,C WHERE C in_scheme CS, CS eid %(cs)s',
                                         'C', self.concept_attributes)
        rset = cnx.execute(concept_rql, {'cs': scheme.eid})
        concepts = tuple(rset.entities(1))
        cache_entities_relations(concepts, rset, 'in_scheme', 'subject')
        scheme._cw_related_cache['in_scheme_object'] = (rset, concepts)
        for rtype in skos.LABELS_RDF_MAPPING:
            label_rql = 'Any L,C WHERE C %s L, C in_scheme CS, CS eid %%(cs)s' % rtype
            label_rql = _select_attributes(label_rql, 'L', self.label_attributes)
            rset = cnx.execute(label_rql, {'cs': scheme.eid})
            cache_entities_relations(concepts, rset, rtype, 'subject')
        for rtype in ('exact_match', 'close_match'):
            rset = cnx.execute('Any X,C,XU WHERE C %s X?, C in_scheme CS, CS eid %%(cs)s, '
                               'X cwuri XU' % rtype, {'cs': scheme.eid})
            cache_entities_relations(concepts, rset, rtype, 'subject')
        for rtype in ('broader_concept', 'related_concept'):
            rset = cnx.execute('Any C1,C2 WHERE C1 %s C2, C1 in_scheme CS, C2 in_scheme CS, '
                               'CS eid %%(cs)s' % rtype, {'cs': scheme.eid})
            cache_entities_relations(concepts, rset, rtype, 'subject', 0, 1)
            if rtype == 'broader_concept':
                cache_entities_relations(concepts, rset, rtype, 'object', 1, 0)


def _select_attributes(rql, var, attributes):
    assert attributes
    select, where = rql.split(' WHERE ')
    for i, attr in enumerate(attributes):
        attrvar = '%s%s' % (var, i)
        where += ', %s %s %s' % (var, attr, attrvar)
        select += ', %s' % attrvar
    return '%s WHERE %s' % (select, where)


def _split_rset(rset, entity_col):
    syntax_tree = rset.syntax_tree()
    syntax_tree.copy = lambda: syntax_tree  # avoid extra copy()
    for entity, subrset in rset.split_rset(col=entity_col, return_dict=True).iteritems():
        subrset._rqlst = syntax_tree
        yield entity, subrset


def cache_entities_relations(entities, rset, rtype, role, entity_col=1, target_col=0):
    """Warm entity relation cache for `rtype` / `role` according to the given rset. This is usually
    useful for multi-valued relations that are not well handled by the ORM by default.

    * `entities` is the list of entities whose cache is expected to be warmed - entities which are
      not found in the `entity_col` column of the rset will have their cache set to an empty rset

    * related entities are in `target_col` column of the rset
    """
    no_relation = set(entities)
    cache_key = '%s_%s' % (rtype, role)
    for entity, subrset in _split_rset(rset, entity_col):
        entity._cw_related_cache[cache_key] = (subrset, tuple(subrset.entities(target_col)))
        try:
            no_relation.remove(entity)
        except KeyError:
            continue
    if no_relation:
        empty_rset = rset.req.empty_rset()
        for entity in no_relation:
            entity._cw_related_cache[cache_key] = (empty_rset, ())



# coding: utf-8
#
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import os
from tempfile import NamedTemporaryFile

from logilab.common.testlib import TestCase, require_module, unittest_main

from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos.rdfio import *
from cubes.skos.rdfio import unicode_with_language as ul


def skos(name):
    return 'http://www.w3.org/2004/02/skos/core#' + name


class UnicodeWithLanguageTC(TestCase):

    def test_comparison_ul_ul(self):
        a = ul('toto', 'fr')
        b = ul('toto', 'fr')
        self.assertEqual(a, b)

        b = ul('toto', 'en')
        self.assertNotEqual(a, b)

        b = ul('titi', 'fr')
        self.assertNotEqual(a, b)

    def test_comparison_ul_and_other(self):
        a = ul('toto', 'fr')
        b = u'toto'
        self.assertNotEqual(a, b)


class GuessRDFFormatTC(TestCase):

    def test_base(self):
        yield self.assertEqual, guess_rdf_format('file.xml'), 'xml'
        yield self.assertEqual, guess_rdf_format('file.rdf'), 'xml'
        yield self.assertEqual, guess_rdf_format('file.n3'), 'n3'

    def test_url(self):
        yield self.assertEqual, guess_rdf_format('http://url.com/rdf/file.xml'), 'xml'
        yield self.assertEqual, guess_rdf_format('http://url.com/rdf/file.rdf'), 'xml'
        yield self.assertEqual, guess_rdf_format('http://url.com/rdf/file.n3'), 'n3'

    def test_error(self):
        yield self.assertRaises, ValueError, guess_rdf_format, 'file.unknown'
        yield self.assertRaises, ValueError, guess_rdf_format, 'http://url.com/rdf/file'
        yield self.assertRaises, ValueError, guess_rdf_format, 'http://url.com/file/rdf'


class RDFRegistryTC(TestCase):
    def setUp(self):
        self.xy = RDFRegistry()
        self.xy.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
        self.xy.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
        self.xy.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')

    def test_register_prefix(self):
        xy = self.xy
        self.assertEqual(xy.prefixes['dc'], 'http://purl.org/dc/elements/1.1/')
        # re-registering the same prefix is fine
        xy.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
        # though re-registering an existing prefix to a different url isn't
        self.assertRaises(RDFRegistryError,
                          xy.register_prefix, 'dc', 'http://purl.org/dc/elements/1.2/')
        # unless we explicitly tell it's ok
        xy.register_prefix('dc', 'http://purl.org/dc/elements/1.2/', overwrite=True)
        self.assertEqual(xy.prefixes['dc'], 'http://purl.org/dc/elements/1.2/')

    def test_register_etype_equivalence(self):
        xy = self.xy
        xy.register_etype_equivalence('CWUser', 'foaf:Person')
        self.assertEqual(xy.etype2rdf['CWUser'], 'http://xmlns.com/foaf/0.1/Person')
        # re-registering the same equivalence is fine
        xy.register_etype_equivalence('CWUser', 'foaf:Person')
        # though re-registering a different isn't
        self.assertRaises(RDFRegistryError,
                          xy.register_etype_equivalence, 'CWUser', 'foaf:Personne')
        # unless we explicitly tell it's ok
        xy.register_etype_equivalence('CWUser', 'foaf:Personne', overwrite=True)
        self.assertEqual(xy.etype2rdf['CWUser'], 'http://xmlns.com/foaf/0.1/Personne')
        xy.register_etype_equivalence('User', 'foaf:Person', overwrite=True)

    def test_register_attribute_equivalence(self):
        xy = self.xy
        xy.register_etype_equivalence('CWUser', 'foaf:Person')
        xy.register_attribute_equivalence('CWUser', 'login', 'dc:title')
        self.assertEqual(xy.attr2rdf[('CWUser', 'login')], 'http://purl.org/dc/elements/1.1/title')
        # re-registering the same equivalence is fine
        xy.register_attribute_equivalence('CWUser', 'login', 'dc:title')
        # though re-registering a different isn't
        self.assertRaises(RDFRegistryError,
                          xy.register_attribute_equivalence, 'CWUser', 'login', 'dc:description')
        # unless we explicitly tell it's ok
        xy.register_attribute_equivalence('CWUser', 'login', 'dc:description', overwrite=True)
        self.assertEqual(xy.attr2rdf[('CWUser', 'login')],
                         'http://purl.org/dc/elements/1.1/description')

    def test_register_relation_equivalence(self):
        xy = self.xy
        xy.register_etype_equivalence('ConceptScheme', 'skos:ConceptScheme')
        xy.register_etype_equivalence('Concept', 'skos:Concept')
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:broader')
        self.assertEqual(xy.rel2rdf[('Concept', 'in_scheme', 'ConceptScheme')],
                         set([('http://www.w3.org/2004/02/skos/core#inScheme', False)]))
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inSchema')
        self.assertEqual(xy.rel2rdf[('Concept', 'in_scheme', 'ConceptScheme')],
                         set([('http://www.w3.org/2004/02/skos/core#inScheme', False),
                              ('http://www.w3.org/2004/02/skos/core#inSchema', False)]))
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:broader')
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:narrower',
                                         reverse=True)
        self.assertEqual(xy.rel2rdf[('Concept', 'broader_concept', 'Concept')],
                         set([('http://www.w3.org/2004/02/skos/core#broader', False),
                              ('http://www.w3.org/2004/02/skos/core#narrower', True)]))

    def test_predicates_for_subject_etype(self):
        xy = self.xy
        xy.register_etype_equivalence('ConceptScheme', 'skos:ConceptScheme')
        xy.register_etype_equivalence('Concept', 'skos:Concept')
        xy.register_attribute_equivalence('ConceptScheme', 'title', 'dc:title')
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:narrower',
                                         reverse=True)
        self.assertEqual(sorted(xy.predicates_for_subject_etype('ConceptScheme')),
                         [('title', 'http://purl.org/dc/elements/1.1/title', False)])
        self.assertEqual(sorted(xy.predicates_for_subject_etype('Concept')),
                         [('broader_concept', 'http://www.w3.org/2004/02/skos/core#narrower', True),
                          ('in_scheme', 'http://www.w3.org/2004/02/skos/core#inScheme', False)])


def dcf(string):
    return u'http://data.culture.fr/thesaurus/resource/ark:/67717/' + string


def ext_entities_to_dict(ext_entities):
    """Turn a sequence of external entities into a 2-level dict with entity
    type and extid as keys.
    """
    entities_dict = {}
    for extentity in ext_entities:
        entities_dict.setdefault(extentity.etype, {})[extentity.extid] = extentity.values
    return entities_dict


class RDFGraphTCMixIn(object):

    def setUp(self):
        graph = self.graph = self.build_graph()
        self.bob = graph.uri("http://example.org/people/Bob")
        self.knows = graph.uri("http://foaf.com/knows")
        self.alice = graph.uri("http://example.org/people/Alice")
        self.firstname = graph.uri("http://foaf.com/firstname")
        self.age = graph.uri("http://foaf.com/age")
        self.desc = graph.uri("http://dc.com/description")
        graph.add(self.bob, self.knows, self.alice)
        graph.add(self.bob, self.firstname, "bob")
        graph.add(self.bob, self.age, 45)
        graph.add(self.bob, self.desc, ul("man", "en"))
        graph.add(self.alice, self.firstname, "alice")
        graph.add(self.alice, self.age, 25)

    def build_graph(self):
        raise NotImplementedError

    _triples = set([
        ('http://example.org/people/Alice', 'http://foaf.com/age', 25),
        ('http://example.org/people/Alice', 'http://foaf.com/firstname', u'alice'),
        ('http://example.org/people/Bob', 'http://foaf.com/firstname', u'bob'),
        ('http://example.org/people/Bob', 'http://foaf.com/age', 45),
        ('http://example.org/people/Bob', 'http://dc.com/description', ul('man', 'en')),
        ('http://example.org/people/Bob', 'http://foaf.com/knows', 'http://example.org/people/Alice'),
    ])

    def test_add(self):
        graph = self.graph
        result = list(graph.objects(self.bob, self.knows))
        self.assertEqual(result, ["http://example.org/people/Alice"])
        result = list(graph.objects(self.bob, self.age))
        self.assertEqual(result, [45])
        result = list(graph.objects(self.bob, self.desc))
        self.assertEqual(result, [ul("man", "en")])
        result = list(graph.objects(self.alice, self.firstname))
        self.assertEqual(result, ["alice"])

    def test_triples(self):
        triples = set(self.graph.triples())
        self.assertEqual(triples, self._triples)

    def test_load_dump_roundtrip(self):
        for rdf_format in ('xml',):  # 'n3', 'nt'): XXX other formats fail, but sounds due to the
                                     # underlying lib
            self.set_description('testing roundtrip for the %s format' % rdf_format)
            yield self.assertRoundTrip, rdf_format

    def assertRoundTrip(self, rdf_format='xml'):
        with NamedTemporaryFile(delete=False) as fobj:
            try:
                self.graph.dump(fobj, rdf_format=rdf_format)
                fobj.close()
                graph = self.build_graph()
                graph.load('file://' + fobj.name, rdf_format=rdf_format)
            finally:
                os.unlink(fobj.name)
        self.assertEqual(set(graph.triples()), self._triples)


class RDFLibRDFGraphTC(RDFGraphTCMixIn, TestCase):
    build_graph = RDFLibRDFGraph

    @require_module('rdflib')
    def setUp(self):
        super(RDFLibRDFGraphTC, self).setUp()


class LibRDFRDFGraphTC(RDFGraphTCMixIn, TestCase):
    build_graph = LibRDFRDFGraph

    @require_module('RDF')
    def setUp(self):
        super(LibRDFRDFGraphTC, self).setUp()


def skos_rdf_registry():
    xy = RDFRegistry()
    xy.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
    xy.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    xy.register_etype_equivalence('ConceptScheme', 'skos:ConceptScheme')
    xy.register_etype_equivalence('Concept', 'skos:Concept')
    xy.register_attribute_equivalence('ConceptScheme', 'title', 'dc:title')
    xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')
    xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:broader')
    xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:narrower',
                                     reverse=True)
    xy.register_relation_equivalence('Concept', 'narrower_concept', 'Concept', 'skos:narrower')
    xy.register_relation_equivalence('Concept', 'narrower_concept', 'Concept', 'skos:broader',
                                     reverse=True)
    xy.register_relation_equivalence('Concept', 'related_concept', 'Concept', 'skos:related')
    xy.register_relation_equivalence('Concept', 'preferred_label', 'Label', 'skos:prefLabel')
    return xy


class RDFGraphToEntitiesTC(TestCase):

    @require_module('rdflib')
    def test_rdflib(self):
        self._test(RDFLibRDFGraph())

    @require_module('RDF')
    def test_librdf(self):
        self._test(LibRDFRDFGraph())

    def _test(self, graph):
        xy = skos_rdf_registry()
        xy.register_relation_equivalence('Concept', 'alternative_label', 'Label', 'skos:altLabel')
        graph.load(self.datapath('siaf_matieres_shortened.xml'))

        etypes = ('ConceptScheme', 'Concept')  # , 'Label')
        ext_entities = rdf_graph_to_entities(xy, graph, etypes)
        e = ext_entities_to_dict(ext_entities)
        self.assertEqual(e['ConceptScheme'][dcf('Matiere')],
                         {'title': set([u"Thésaurus-matières pour l'indexation des archives locales"])})
        self.assertEqual(e['Concept'][dcf('T1-1073')],
                         {'broader_concept': set([dcf('T1-3')]),
                          'in_scheme': set([dcf('Matiere')]),
                          'related_concept': set([dcf('T1-760')]),
                          'alternative_label': set([ul(u'académie', 'fr-fr'),
                                                    ul(u"société d'agriculture", 'fr-fr')]),
                          'preferred_label': set([ul(u'société savante', 'fr-fr')]),
                          })


class AddEntitiesToRDFGraphTC(CubicWebTC):

    @require_module('rdflib')
    def test_rdflib(self):
        self._test(RDFLibRDFGraph())

    @require_module('RDF')
    def test_librdf(self):
        self._test(LibRDFRDFGraph())

    def _test(self, graph):
        xy = skos_rdf_registry()
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'some classification')
            c1 = cnx.create_entity('Concept', in_scheme=scheme)
            l1 = cnx.create_entity('Label', label=u'hip', language_code=u'fr',
                                   kind=u'preferred', label_of=c1)
            c2 = cnx.create_entity('Concept', in_scheme=scheme, broader_concept=c1)
            l2 = cnx.create_entity('Label', label=u'hop', language_code=u'fr',
                                   kind=u'preferred', label_of=c2)
            cnx.commit()
            RDFGraphGenerator(graph, GraphWalker([scheme], accept=lambda x: True)).add_entities(xy)
            expected = [
                (scheme.absolute_url(), TYPE_PREDICATE_URI, skos('ConceptScheme')),
                (scheme.absolute_url(), 'http://purl.org/dc/elements/1.1/title', u'some classification'),
                (c1.absolute_url(), skos('prefLabel'), l1.absolute_url()),
                (c2.absolute_url(), skos('prefLabel'), l2.absolute_url()),
                (c1.absolute_url(), skos('narrower'), c2.absolute_url()),
                (c1.absolute_url(), TYPE_PREDICATE_URI, skos('Concept')),
                (c1.absolute_url(), skos('inScheme'), scheme.absolute_url()),
                (c2.absolute_url(), TYPE_PREDICATE_URI, skos('Concept')),
                (c2.absolute_url(), skos('inScheme'), scheme.absolute_url()),
                (c2.absolute_url(), skos('broader'), c1.absolute_url()),
            ]
            self.assertCountEqual(graph.triples(), expected)

if __name__ == '__main__':
    unittest_main()



from cubicweb.dataimport import ucsvreader
from cubes.skos import rdfio
from cubes.skos.rdfio import unicode_with_language as ul

_ = unicode


class InvalidLCSVFile(RuntimeError):
    """LCSV input is malformed."""

    def __init__(self, msg, column=None):
        super(InvalidLCSVFile, self).__init__(msg)
        self.column = column


class LCSV2RDF(object):
    """Create RDF triples expressing the data contained in the LCSV stream
    reference : W3C unofficial draft, http://jenit.github.io/linked-csv/
    """

    def __init__(self, stream, delimiter, encoding, uri_generator, uri_cls, default_lang=None):
        """ check stream validity and init attribute"""
        self.source_file = ucsvreader(stream, encoding=encoding, delimiter=delimiter)
        # raise an error value if the line doesn't contains a #
        self.uri_gen = uri_generator
        self.prefixes = {'skos': 'http://www.w3.org/2004/02/skos/core#',
                         'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'}
        self.uri_func = lambda string: uri_cls(rdfio.normalize_uri(string, self.prefixes))
        self.uris = {}
        self.no_information_missing = None
        self.prolog = {}
        self.default_lang = default_lang
        header = self.source_file.next()
        try:
            self.prolog_column = header.index('#')
        except ValueError:
            raise InvalidLCSVFile(_("missing prolog column (#)"))
        try:
            self.id_column = header.index('$id')
        except ValueError:
            raise InvalidLCSVFile(_("missing $id column"))

    def triples(self):
        """Yield triples from the lcsv stream"""
        for line in self.source_file:
            # prolog lines must all be at the start of a linked CSV file (dixit draft W3C)
            end_of_prolog = False
            if line[self.prolog_column] and not end_of_prolog:
                self.parse_prolog_line(line)
            else:
                end_of_prolog = True
                self.check_no_information_missing()
                for triple in self.parse_data_line(line):
                    yield triple

    def check_no_information_missing(self):
        """ raise an InvalidLCSVFile error if necessary information are missing"""
        if not self.no_information_missing:
            for column, info in self.prolog.items():
                if not info.get('url', None):
                    raise InvalidLCSVFile(_("missing url info for column %s"), column=str(column))
            self.no_information_missing = True

    def parse_prolog_line(self, line):
        """analyse and store the prolog informations (url, type) described in this line"""
        prolog_type = line[self.prolog_column]
        for index, value in enumerate(line):
            if index != self.prolog_column and value:
                self.prolog.setdefault(index, {})[prolog_type] = value

    def parse_data_line(self, line):
        """analyse the lcsv line and yield the resulting triples"""
        rdftype = self.uri_func(self.prolog[self.id_column]['url'])
        subj = self.uri_func(self.uri_gen(line[self.id_column]))
        yield (subj, self.uri_func('rdf:type'), rdftype)
        self.uris[line[self.id_column]] = subj
        for column_index, column_info in self.prolog.items():
            if column_index != self.id_column and line[column_index]:
                predicate = self.uri_func(column_info['url'])
                content = line[column_index].strip()
                obj = self.type_object(content, column_info)
                yield (subj, predicate, obj)

    def type_object(self, obj, column_info):
        """return the object of the triple considering the value and the prolog infos given"""
        obj_type = column_info.get('type', 'string')
        if obj_type == 'url':
            try:
                obj = self.uris[obj]
            except KeyError:
                obj = self.uri_func(obj)
        elif obj_type == 'string':
            lang = column_info.get('lang', self.default_lang)
            obj = ul(obj, lang) if lang else obj
        elif obj_type == 'integer':
            obj = int(obj)
        return obj



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos rdf views"""

from io import BytesIO

from cubicweb.__pkginfo__ import numversion
from cubicweb.predicates import adaptable
from cubicweb.view import EntityView

from cubes.skos.rdfio import default_graph


class RDFPrimaryView(EntityView):
    """RDF primary view outputting complete information of possibly several
    entities.
    """
    __regid__ = 'primary.rdf'
    adapter = 'RDFPrimary'
    __select__ = EntityView.__select__ & adaptable(adapter)
    templatable = False
    content_type = 'text/xml'
    binary = numversion >= (3, 21)

    def call(self):
        graph = default_graph()
        for entity in self.cw_rset.entities():
            self.entity_call(entity, graph)
        self._dump(graph)

    def entity_call(self, entity, graph=None):
        dump = graph is None
        if graph is None:
            graph = default_graph()
        rdfgenerator = entity.cw_adapt_to(self.adapter)
        rdfgenerator.fill(graph)
        if dump:
            self._dump(graph)

    if binary:

        def _dump(self, graph):
            fobj = BytesIO()
            graph.dump(fobj, rdf_format='xml')
            self.w(fobj.getvalue())

    else:

        def _dump(self, graph):
            fobj = BytesIO()
            graph.dump(fobj, rdf_format='xml')
            self.w(unicode(fobj.getvalue(), 'utf-8'))


class RDFListView(RDFPrimaryView):
    """RDF list view with minimal information on each entity."""
    __regid__ = 'list.rdf'
    adapter = 'RDFList'
    __select__ = EntityView.__select__ & adaptable(adapter)



add_attribute('Label', 'kind')
add_relation_definition('Label', 'label_of', 'Concept')
for kind, rtype in [('preferred', 'pref_label_of'),
                    ('alternative', 'alt_label_of'),
                    ('hidden', 'hidden_label_of')]:
    sql("UPDATE cw_Label SET cw_kind='%s', cw_label_of=cw_%s WHERE NOT cw_%s IS NULL"
        % (kind, rtype, rtype))
    sql("UPDATE cw_Label SET cw_%s=NULL" % rtype)
    commit()
    drop_relation_type(rtype)
sync_schema_props_perms('Label')  # drop unique together constraint

# renaming of computed relation type is broken, drop then add
drop_relation_type('pref_label')
add_relation_type('preferred_label')
drop_relation_type('alt_label')
add_relation_type('alternative_label')
sync_schema_props_perms('hidden_label')



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos hooks"""

from yams import ValidationError

from cubicweb.server import hook
from cubicweb.predicates import is_instance

from cubes.skos.sobjects import dump_relations

_ = unicode


class CreateRelationsOp(hook.DataOperationMixIn, hook.Operation):
    """Data operation updating the modification date of its data entities."""
    containercls = list

    def precommit_event(self):
        cnx = self.cnx
        for eid, relations in self.get_data():
            for subject_eid, rtype, object_eid in relations:
                if subject_eid is None:
                    subject_eid = eid
                elif cnx.deleted_in_transaction(subject_eid):
                    continue
                if object_eid is None:
                    object_eid = eid
                elif cnx.deleted_in_transaction(object_eid):
                    continue
                cnx.execute('SET X %s Y WHERE X eid %%(x)s, Y eid %%(y)s' % rtype,
                            {'x': subject_eid, 'y': object_eid})


class ReplaceExternalUriByEntityHook(hook.Hook):
    """Abstract hook to replace an ExternalUri with the same cwuri as the selected entity. To
    activate this behaviour, simply inherit from this hook and set a proper selector.

    Notice the hook has to be done on a 'before_add_entity' event as in case where the entity and
    the external uri are from an external source they will as such usually share the same extid
    which is a unique column in the `entities` system table. So this hook looks for a matching
    external uri, eventually records its relations into a data operation for later insertion
    and finally removes it.
    """
    __abstract__ = True
    __regid__ = 'skos.replace_externaluri'
    events = ('before_add_entity',)

    def __call__(self):
        cnx = self._cw
        rset = cnx.execute('ExternalUri X WHERE X cwuri %(uri)s', {'uri': self.entity.cwuri})
        if rset:
            exturi_eid = rset[0][0]
            # remember the external uri's relations and store them in an operation
            relations = dump_relations(cnx, exturi_eid, 'ExternalUri')
            CreateRelationsOp.get_instance(self._cw).add_data((self.entity.eid, relations))
            # delete the external uri before insertion of the new entity
            cnx.execute('DELETE ExternalUri X WHERE X eid %(x)s', {'x': exturi_eid})


class ReplaceExternalUriByConceptHook(ReplaceExternalUriByEntityHook):
    """Replace ExternalUri by a Concept"""
    __select__ = ReplaceExternalUriByEntityHook.__select__ & is_instance('Concept')


# SKOSSource <-> CWSource synchronization ######################################

# remove once fix for https://www.cubicweb.org/ticket/5477315 is published
from cubicweb import server  # noqa
server.BEFORE_ADD_RELATIONS.add('through_cw_source')


class CreateCWSource(hook.Hook):
    """Create a CWSource upon creation of a SKOSSource"""
    __regid__ = 'skos.create-source'
    __select__ = hook.Hook.__select__ & is_instance('SKOSSource')
    events = ('before_add_entity',)

    def __call__(self):
        if 'through_cw_source' in self.entity.cw_edited:
            return
        source = self._cw.create_entity('CWSource', name=self.entity.name, url=self.entity.url,
                                        type=u'datafeed', parser=u'rdf.skos',
                                        config=u'synchronize=no\nuse-cwuri-as-url=no')
        self.entity.cw_edited['through_cw_source'] = source.eid
        # remove once fix for https://www.cubicweb.org/ticket/5477315 is published
        nocheck = self._cw.transaction_data.setdefault('skip-security', set())
        nocheck.add((self.entity.eid, 'through_cw_source', source.eid))


class UpdateCWSource(hook.Hook):
    """update associated CWSource upon modification of name or url of a SKOSSource"""
    __regid__ = 'skos.update-source'
    __select__ = hook.Hook.__select__ & is_instance('SKOSSource')
    events = ('after_update_entity',)

    def __call__(self):
        edited = self.entity.cw_edited
        to_update = {}
        if 'name' in edited:
            to_update['name'] = edited['name']
        if 'url' in edited:
            to_update['url'] = edited['url']
        if to_update:
            self.entity.through_cw_source[0].cw_set(**to_update)


# Ensure Concept has a single preferred label in a given language label, kind, lang ################

class CheckPreferredLabelOp(hook.DataOperationMixIn, hook.Operation):
    """Data operation checking non-redundancy of labels of kind "preferred"."""
    containercls = list

    def precommit_event(self):
        concept_labels = dict((label.label_of[0].eid, label) for label in self.get_data())
        concept_eids = ',' . join(unicode(concept_eid) for concept_eid in concept_labels)
        rql = ('Any C WHERE EXISTS('
               'C preferred_label L, L language_code LC, '
               'C preferred_label L2, L2 language_code LC, '
               'NOT L identity L2), C eid IN (%s)' % concept_eids)
        for eid, in self.cnx.execute(rql):
            errors = {'': _('a preferred label in "%(lang)s" language already exists'),
                      'language_code': _('please use another language code')}
            label = concept_labels[eid]
            raise ValidationError(label.eid, errors, {'lang': label.language_code})
        rql = ('Any C WHERE EXISTS('
               'C preferred_label L, L language_code NULL, '
               'C preferred_label L2, L2 language_code NULL, '
               'NOT L identity L2), C eid IN (%s)' % concept_eids)
        for eid, in self.cnx.execute(rql):
            errors = {'': _('a preferred label without language already exists'),
                      'language_code': _('please specify a language code')}
            label = concept_labels[eid]
            raise ValidationError(label.eid, errors)


class CreateOrUpdatePrefLabelHook(hook.Hook):
    """Check for duplicate definition of a Label of kind "preferred"."""
    __regid__ = 'skos.check-preferred-label'
    __select__ = hook.Hook.__select__ & is_instance('Label')
    events = ('after_add_entity', 'after_update_entity')

    def __call__(self):
        edited = self.entity.cw_edited
        if edited.get('kind') == 'preferred':
            CheckPreferredLabelOp.get_instance(self._cw).add_data(self.entity)
        elif 'language_code' in edited and self.entity.kind == 'preferred':
            CheckPreferredLabelOp.get_instance(self._cw).add_data(self.entity)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
from itertools import imap

from logilab.mtconverter import xml_escape


def cwuri2eid(cnx, etypes, source_eid=None):
    """return a dictionary mapping cwuri to eid for already imported entities"""
    rql = 'Any U, X WHERE X cwuri U'
    args = {}
    if len(etypes) == 1:
        rql += ', X is %s' % etypes[0]
    else:
        assert etypes, 'no entity types specified'
        rql += ', X is IN (%s)' % ','.join(etypes)
    if source_eid is not None:
        rql += ', X cw_source S, S eid %(s)s'
        args['s'] = source_eid
    return dict(cnx.execute(rql, args))


class ExtEntitiesImporter(object):
    """This class is responsible for importing a transitional representation
    of entities coming from an external source. These external entities are
    expected to be instances of `ExtEntity` built from the entity type, the
    extid (as a string) and a dict mapping attributes and subject relations to
    the set of corresponding values.

    ::

        ExtEntity(etype='Label',
                  extid='http://example.org/ark:/000000/123',
                  values={'label': set([u'toto']),
                          'pref_label_of': set(['http://example.org/ark:/000000/456']),
                         }
                 )
    """

    def __init__(self, cnx, store, import_log, source=None,
                 raise_on_error=False, etypes_order_hint=(), extid2eid=None):
        self.cnx = cnx
        self.store = store
        self.import_log = import_log
        if source is None:
            source = self.cnx.repo.system_source
        self.source = source
        self.raise_on_error = raise_on_error
        self.etypes_order_hint = etypes_order_hint
        # index of already imported entities
        if extid2eid is None:
            extid2eid = cwuri2eid(cnx, etypes_order_hint, source.eid)
        self.extid2eid = extid2eid

    def import_entities(self, entities, use_extid_as_cwuri=True, skip_known_external_uris=True):
        """Import external entities given as ``ExtEntity``s.

        Parameters:

        * `use_extid_as_cwuri` use external entity's extid as cwuri (True by default).

        * `skip_known_external_uris` skip ExternalUri entities which are already known (True by
          default).

        """
        if use_extid_as_cwuri:
            def set_cwuri_if_needed(extentity, extid2eid=self.extid2eid):
                """Use external entity's extid as cwuri"""
                if extentity.extid not in extid2eid:
                    extentity.values['cwuri'] = set([unicode(extentity.extid)])
                return extentity
            entities = imap(set_cwuri_if_needed, entities)
        if skip_known_external_uris:
            entities = (ee for ee in entities
                        if not (ee.etype == 'ExternalUri' and ee.extid in self.extid2eid))
        # {etype: [etype dict]} of entities that are in the import queue
        queue = {}
        # set of created/updated eids
        self.created = set()
        self.updated = set()
        # order entity dictionaries then create/update them
        deferred = self._import_entities(entities, queue)
        # create deferred relations that don't exist already
        missing_relations = self.create_deferred_relations(deferred)
        self._warn_about_missing_work(queue, missing_relations)
        return self.created, self.updated

    def _import_entities(self, entities, queue):
        entity_from_eid = self.cnx.entity_from_eid
        extid2eid = self.extid2eid
        deferred = {}  # non inlined relations that may be deferred
        self.import_log.record_debug('importing entities')
        for extentity in self.iter_ext_entities(entities, deferred, queue):
            try:
                entity = entity_from_eid(extid2eid[extentity.extid], extentity.etype)
                assert extentity.etype == entity.cw_etype
            except KeyError:
                self.create_entity(extentity).eid
            else:
                if extentity.values:
                    # XXX some inlined relations may already exists
                    entity.cw_set(**extentity.values)
                    self.updated.add(entity.eid)
        return deferred

    def create_entity(self, extentity):
        entity = self.store.create_entity(extentity.etype, **extentity.values)
        self.extid2eid[extentity.extid] = entity.eid
        self.created.add(entity.eid)
        return entity

    def create_deferred_relations(self, deferred):
        relate = self.store.relate
        rschema = self.cnx.vreg.schema.rschema
        extid2eid = self.extid2eid
        missing_relations = []
        for rtype, relations in deferred.items():
            self.import_log.record_debug('importing %s %s relations' % (len(relations), rtype))
            symmetric = rschema(rtype).symmetric
            existing = self.existing_relations(rtype)
            for subject_uri, object_uri in relations:
                try:
                    subject_eid = extid2eid[subject_uri]
                    object_eid = extid2eid[object_uri]
                except KeyError:
                    missing_relations.append((subject_uri, rtype, object_uri))
                    continue
                if (subject_eid, object_eid) not in existing:
                    relate(subject_eid, rtype, object_eid)
                    existing.add((subject_eid, object_eid))
                    if symmetric:
                        existing.add((object_eid, subject_eid))
        return missing_relations

    def _warn_about_missing_work(self, queue, missing_relations):
        error = self.import_log.record_error
        if queue:
            msgs = ["can't create some entities, is there some cycle or "
                    "missing data?"]
            for etype, ext_entities in queue.items():
                for extentity in ext_entities:
                    msgs.append(str(extentity))
            map(error, msgs)
            if self.raise_on_error:
                raise Exception('\n'.join(msgs))
        if missing_relations:
            msgs = ["can't create some relations, is there missing data?"]
            for subject_uri, rtype, object_uri in missing_relations:
                msgs.append("%s %s %s" % (subject_uri, rtype, object_uri))
            map(error, msgs)
            if self.raise_on_error:
                raise Exception('\n'.join(msgs))

    def existing_relations(self, rtype):
        """return a set of (subject, object) eids already related by `rtype`"""
        rql = 'Any X,O WHERE X %s O, X cw_source S, O cw_source S, S eid %%(s)s' % rtype
        return set(tuple(x) for x in self.cnx.execute(rql, {'s': self.source.eid}))

    def iter_ext_entities(self, entities, deferred, queue):
        """Yield external entities in an order which attempts to satisfy
        schema constraints (inlined / cardinality) and to optimize the import.
        """
        schema = self.cnx.vreg.schema
        extid2eid = self.extid2eid
        for extentity in entities:
            # check data in the transitional representation and prepare it for
            # later insertion in the database
            for subject_uri, rtype, object_uri in extentity.prepare(schema):
                deferred.setdefault(rtype, set()).add((subject_uri, object_uri))
            if not extentity.is_ready(extid2eid):
                queue.setdefault(extentity.etype, []).append(extentity)
                continue
            yield extentity
            # check for some entities in the que that may now be ready
            for etype in self.etypes_order_hint:
                if etype in queue:
                    new_queue = []
                    for extentity in queue[etype]:
                        if extentity.is_ready(extid2eid):
                            yield extentity
                        else:
                            new_queue.append(extentity)
                    if new_queue:
                        queue[etype][:] = new_queue
                    else:
                        del queue[etype]


class SimpleImportLog(object):
    """Fake CWDataImport log using a simple text format.

    Useful to display logs in the UI instead of storing them to the
    database.
    """

    def __init__(self, filename):
        self.logs = []
        self.filename = filename

    def record_debug(self, msg, path=None, line=None):
        self._log(logging.DEBUG, msg, path, line)

    def record_info(self, msg, path=None, line=None):
        self._log(logging.INFO, msg, path, line)

    def record_warning(self, msg, path=None, line=None):
        self._log(logging.WARNING, msg, path, line)

    def record_error(self, msg, path=None, line=None):
        self._log(logging.ERROR, msg, path, line)

    def record_fatal(self, msg, path=None, line=None):
        self._log(logging.FATAL, msg, path, line)

    def _log(self, severity, msg, path, line):
        encodedmsg = u'%s\t%s\t%s\t%s' % (severity, self.filename,
                                          line or u'', msg)
        self.logs.append(encodedmsg)


class HTMLImportLog(SimpleImportLog):
    """Fake CWDataImport log using a simple HTML format."""

    def _log(self, severity, msg, path, line):
        encodedmsg = u'%s\t%s\t%s\t%s<br/>' % (severity, self.filename,
                                               line or u'', xml_escape(msg))
        self.logs.append(encodedmsg)
